# -*- coding: utf-8 -*- #
# Copyright 2013 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""One-line documentation for auth module.

A detailed description of auth.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import contextlib
import datetime
import json
import os
import textwrap
import time

import dateutil
from google.auth import exceptions as google_auth_exceptions
from google.auth import external_account as google_auth_external_account
import google.auth.compute_engine as google_auth_gce
from googlecloudsdk.api_lib.auth import util as auth_util
from googlecloudsdk.core import config
from googlecloudsdk.core import context_aware
from googlecloudsdk.core import http
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core import requests
from googlecloudsdk.core import transport
from googlecloudsdk.core.configurations import named_configs
from googlecloudsdk.core.credentials import creds as c_creds
from googlecloudsdk.core.credentials import exceptions as creds_exceptions
from googlecloudsdk.core.credentials import gce as c_gce
from googlecloudsdk.core.util import encoding
from googlecloudsdk.core.util import files
from googlecloudsdk.core.util import times
import httplib2
from oauth2client import client
from oauth2client import crypt
from oauth2client import service_account
from oauth2client.contrib import gce as oauth2client_gce
from oauth2client.contrib import reauth_errors
import six
from six.moves import urllib


ACCESS_TOKEN_ENV_VAR_NAME = (
    'CLOUDSDK_AUTH_ACCESS_TOKEN')
GOOGLE_OAUTH2_PROVIDER_AUTHORIZATION_URI = (
    'https://accounts.google.com/o/oauth2/auth')
GOOGLE_OAUTH2_PROVIDER_REVOKE_URI = (
    'https://accounts.google.com/o/oauth2/revoke')
_GRANT_TYPE = 'urn:ietf:params:oauth:grant-type:jwt-bearer'

_CREDENTIALS_EXPIRY_WINDOW = '300s'


class Error(creds_exceptions.Error):
  """Exceptions for the credentials module."""


class NoImpersonationAccountError(Error):
  """Exception when there is no account to impersonate."""


class PrintTokenAuthenticationException(Error):
  """Exceptions that tell the users to run auth login."""

  def __init__(self, message):
    super(PrintTokenAuthenticationException, self).__init__(textwrap.dedent("""\
        {message}
        Please run:

          $ gcloud auth login

        to obtain new credentials.

        For service account, please activate it first:

          $ gcloud auth activate-service-account ACCOUNT""".format(
              message=message)))


class NoCredentialsForAccountException(PrintTokenAuthenticationException):
  """Exception for when no credentials are found for an account."""

  def __init__(self, account):
    super(NoCredentialsForAccountException, self).__init__(
        'Your current active account [{account}] does not have any'
        ' valid credentials'.format(account=account))


class InvalidCredentialFileException(Error):
  """Exception for when an external credential file could not be loaded."""

  def __init__(self, f, e):
    super(InvalidCredentialFileException, self).__init__(
        'Failed to load credential file: [{f}].  {message}'
        .format(f=f, message=six.text_type(e)))


class AccountImpersonationError(Error):
  """Exception for when attempting to impersonate a service account fails."""
  pass


class FlowError(Error):
  """Exception for when something goes wrong with a web flow."""


class RevokeError(Error):
  """Exception for when there was a problem revoking."""


class InvalidCodeVerifierError(Error):
  """Exception for invalid code verifier for pkce."""


class UnsupportedCredentialsError(Error):
  """Exception for when a credential type is not supported."""


IMPERSONATION_TOKEN_PROVIDER = None


class StaticCredentialProviders(object):
  """Manages a list of credential providers."""

  def __init__(self):
    self._providers = []

  def AddProvider(self, provider):
    self._providers.append(provider)

  def RemoveProvider(self, provider):
    self._providers.remove(provider)

  def GetCredentials(self, account, use_google_auth=True):
    for provider in self._providers:
      cred = provider.GetCredentials(account, use_google_auth)
      if cred is not None:
        return cred
    return None

  def GetAccounts(self):
    accounts = set()
    for provider in self._providers:
      accounts |= provider.GetAccounts()
    return accounts


STATIC_CREDENTIAL_PROVIDERS = StaticCredentialProviders()


class GceCredentialProvider(object):
  """Provides account, project and credential data for gce vm env."""

  def GetCredentials(self, account, use_google_auth=True):
    if account in c_gce.Metadata().Accounts():
      return AcquireFromGCE(account, use_google_auth)
    return None

  def GetAccount(self):
    if properties.VALUES.core.check_gce_metadata.GetBool():
      return c_gce.Metadata().DefaultAccount()
    return None

  def GetAccounts(self):
    return set(c_gce.Metadata().Accounts())

  def GetProject(self):
    if properties.VALUES.core.check_gce_metadata.GetBool():
      return c_gce.Metadata().Project()
    return None

  def Register(self):
    properties.VALUES.core.account.AddCallback(self.GetAccount)
    properties.VALUES.core.project.AddCallback(self.GetProject)
    STATIC_CREDENTIAL_PROVIDERS.AddProvider(self)

  def UnRegister(self):
    properties.VALUES.core.account.RemoveCallback(self.GetAccount)
    properties.VALUES.core.project.RemoveCallback(self.GetProject)
    STATIC_CREDENTIAL_PROVIDERS.RemoveProvider(self)


def AvailableAccounts():
  """Get all accounts that have credentials stored for the CloudSDK.

  This function will also ping the GCE metadata server to see if GCE credentials
  are available.

  Returns:
    [str], List of the accounts.

  """
  store = c_creds.GetCredentialStore()
  accounts = store.GetAccounts() | STATIC_CREDENTIAL_PROVIDERS.GetAccounts()

  return sorted(accounts)


def GoogleAuthDisabledGlobally():
  """Returns True if google-auth is disabled globally."""
  return properties.VALUES.auth.disable_load_google_auth.GetBool()


def _TokenExpiresWithinWindow(expiry_window,
                              token_expiry_time,
                              max_window_seconds=3600):
  """Determines if token_expiry_time is within expiry_window_duration.

  Calculates the amount of time between utcnow() and token_expiry_time and
  returns true, if that amount is less than the provided duration window. All
  calculations are done in number of seconds for consistency.


  Args:
    expiry_window: string, Duration representing the amount of time between
      now and token_expiry_time to compare against.
    token_expiry_time: datetime, The time when token expires.
    max_window_seconds: int, Maximum size of expiry window, in seconds.

  Raises:
    ValueError: If expiry_window is invalid or can not be parsed.

  Returns:
    True if token is expired or will expire with in the provided window,
    False otherwise.
  """
  try:
    min_expiry = times.ParseDuration(expiry_window, default_suffix='s')
    if min_expiry.total_seconds > max_window_seconds:
      raise ValueError('Invalid expiry window duration [{}]: '
                       'Must be between 0s and 1h'.format(expiry_window))
  except times.Error as e:
    message = six.text_type(e).rstrip('.')
    raise ValueError('Error Parsing expiry window duration '
                     '[{}]: {}'.format(expiry_window, message))

  token_expiry_time = times.LocalizeDateTime(token_expiry_time,
                                             tzinfo=dateutil.tz.tzutc())
  window_end = times.GetDateTimePlusDuration(
      times.Now(tzinfo=dateutil.tz.tzutc()), min_expiry)

  return token_expiry_time <= window_end


def _GetAccessTokenFromCreds(creds):
  if creds is None:
    return None
  if c_creds.IsGoogleAuthCredentials(creds):
    return creds.token
  else:
    return creds.access_token


def GetAccessToken(account=None, scopes=None, allow_account_impersonation=True):
  """Returns the access token of the given account or the active account.

  GetAccessToken ignores whether credentials have been disabled via properties.
  Use this function when the caller absolutely requires credentials.

  Args:
    account: str, The account to get the access token for. If None, the
      account stored in the core.account property is used.
    scopes: tuple, Custom auth scopes to request. By default CLOUDSDK_SCOPES are
      requested.
    allow_account_impersonation: bool, True to allow use of impersonated service
      account credentials (if that is configured).
  """
  creds = Load(account, scopes, False, allow_account_impersonation, True)
  return _GetAccessTokenFromCreds(creds)


def GetAccessTokenIfEnabled(account=None,
                            scopes=None,
                            allow_account_impersonation=True):
  """Returns the access token of the given account or the active account.

  If credentials have been disabled via properties, this will return None.
  Otherwise it return the access token of the account like normal. Use this
  function when credentials are optional for the caller, or the caller want to
  handle the situation of credentials being disabled by properties.

  Args:
    account: str, The account to get the access token for. If None, the
      account stored in the core.account property is used.
    scopes: tuple, Custom auth scopes to request. By default CLOUDSDK_SCOPES are
      requested.
    allow_account_impersonation: bool, True to allow use of impersonated service
      account credentials (if that is configured).
  """
  if properties.VALUES.auth.disable_credentials.GetBool():
    return None
  return GetAccessToken(account, scopes, allow_account_impersonation)


def GetFreshAccessToken(account=None,
                        scopes=None,
                        min_expiry_duration='1h',
                        allow_account_impersonation=True):
  """Returns a fresh access token of the given account or the active account.

  Same as GetAccessToken except that the access token returned by
  this function is valid for at least min_expiry_duration.

  Args:
    account: str, The account to get the access token for. If None, the
      account stored in the core.account property is used.
    scopes: tuple, Custom auth scopes to request. By default CLOUDSDK_SCOPES are
      requested.
    min_expiry_duration: Duration str, Refresh the token if they are
      within this duration from expiration. Must be a valid duration between 0
      seconds and 1 hour (e.g. '0s' >x< '1h').
    allow_account_impersonation: bool, True to allow use of impersonated service
      account credentials (if that is configured).
  """
  creds = LoadFreshCredential(account, scopes, min_expiry_duration,
                              allow_account_impersonation, True)
  return _GetAccessTokenFromCreds(creds)


def GetFreshAccessTokenIfEnabled(account=None,
                                 scopes=None,
                                 min_expiry_duration='1h',
                                 allow_account_impersonation=True):
  """Returns a fresh access token of the given account or the active account.

  Same as GetAccessTokenIfEnabled except that the access token returned by
  this function is valid for at least min_expiry_duration.

  Args:
    account: str, The account to get the access token for. If None, the
      account stored in the core.account property is used.
    scopes: tuple, Custom auth scopes to request. By default CLOUDSDK_SCOPES are
      requested.
    min_expiry_duration: Duration str, Refresh the token if they are
      within this duration from expiration. Must be a valid duration between 0
      seconds and 1 hour (e.g. '0s' >x< '1h').
    allow_account_impersonation: bool, True to allow use of impersonated service
      account credentials (if that is configured).
  """
  if properties.VALUES.auth.disable_credentials.GetBool():
    return None
  return GetFreshAccessToken(account, scopes, min_expiry_duration,
                             allow_account_impersonation)


def LoadFreshCredential(account=None,
                        scopes=None,
                        min_expiry_duration='1h',
                        allow_account_impersonation=True,
                        use_google_auth=True):
  """Load credentials and force a refresh.

    Will always refresh loaded credential if it is expired or would expire
    within min_expiry_duration.

  Args:
    account: str, The account address for the credentials being fetched. If
      None, the account stored in the core.account property is used.
    scopes: tuple, Custom auth scopes to request. By default CLOUDSDK_SCOPES are
      requested.
    min_expiry_duration: Duration str, Refresh the credentials if they are
      within this duration from expiration. Must be a valid duration between 0
      seconds and 1 hour (e.g. '0s' >x< '1h').
    allow_account_impersonation: bool, True to allow use of impersonated service
      account credentials (if that is configured). If False, the active user
      credentials will always be loaded.
    use_google_auth: bool, True to load credentials as google-auth credentials.
      False to load credentials as oauth2client credentials..

  Returns:
    oauth2client.client.Credentials or google.auth.credentials.Credentials.
    When all of the following conditions are met, it returns
    google.auth.credentials.Credentials and otherwise it returns
    oauth2client.client.Credentials.

    * use_google_auth is True
    * google-auth is not globally disabled by auth/disable_load_google_auth.

  Raises:
    NoActiveAccountException: If account is not provided and there is no
        active account.
    NoCredentialsForAccountException: If there are no valid credentials
        available for the provided or active account.
    c_gce.CannotConnectToMetadataServerException: If the metadata server cannot
        be reached.
    TokenRefreshError: If the credentials fail to refresh.
    TokenRefreshReauthError: If the credentials fail to refresh due to reauth.
    AccountImpersonationError: If impersonation is requested but an
      impersonation provider is not configured.
   ValueError:
  """
  cred = Load(
      account=account,
      scopes=scopes,
      allow_account_impersonation=allow_account_impersonation,
      use_google_auth=use_google_auth)
  RefreshIfExpireWithinWindow(cred, min_expiry_duration)

  return cred


def LoadIfEnabled(allow_account_impersonation=True, use_google_auth=True):
  """Get the credentials associated with the current account.

  If credentials have been disabled via properties, this will return None.
  Otherwise it will load credentials like normal. If credential loading fails
  for any reason (including the user not being logged in), the usual exception
  is raised.

  Args:
    allow_account_impersonation: bool, True to allow use of impersonated service
      account credentials (if that is configured). If False, the active user
      credentials will always be loaded.
    use_google_auth: bool, True to load credentials as google-auth credentials.
    False to load credentials as oauth2client credentials..

  Returns:
    oauth2client.client.Credentials or google.auth.credentials.Credentials if
    credentials are enabled. When all of the following conditions are met, it
    returns google.auth.credentials.Credentials and otherwise it returns
    oauth2client.client.Credentials.

    * use_google_auth is True
    * google-auth is not globally disabled by auth/disable_load_google_auth.

  Raises:
    NoActiveAccountException: If account is not provided and there is no
        active account.
    c_gce.CannotConnectToMetadataServerException: If the metadata server cannot
        be reached.
    TokenRefreshError: If the credentials fail to refresh.
    TokenRefreshReauthError: If the credentials fail to refresh due to reauth.
  """
  if properties.VALUES.auth.disable_credentials.GetBool():
    return None
  return Load(
      allow_account_impersonation=allow_account_impersonation,
      use_google_auth=use_google_auth)


def ParseImpersonationAccounts(service_account_ids):
  """Finds the target impersonation principal and the delegates.

  Args:
     service_account_ids: str, A list of service account ids separated using
       comma.

  Returns:
     A tuple (target_principal, delegates).

  Raises:
    NoImpersonationAccountError: if the input does not contain service accounts.
  """
  service_account_ids = service_account_ids.split(',')
  service_account_ids = [sa_id.strip() for sa_id in service_account_ids]
  if not service_account_ids:
    raise NoImpersonationAccountError('No service account to impersonate.')
  return service_account_ids[-1], (service_account_ids[:-1] or None)


def Load(account=None,
         scopes=None,
         prevent_refresh=False,
         allow_account_impersonation=True,
         use_google_auth=True,
         with_access_token_cache=True):
  """Get the credentials associated with the provided account.

  This loads credentials regardless of whether credentials have been disabled
  via properties. Only use this when the functionality of the caller absolutely
  requires credentials (like printing out a token) vs logically requiring
  credentials (like for an http request).

  Credential information may come from the stored credential file (representing
  the last gcloud auth command), or the credential cache (representing the last
  time the credentials were refreshed). If they come from the cache, the
  token_response field will be None, as the full server response from the cached
  request was not stored.

  Args:
    account: str, The account address for the credentials being fetched. If
        None, the account stored in the core.account property is used.
    scopes: tuple, Custom auth scopes to request. By default CLOUDSDK_SCOPES
        are requested.
    prevent_refresh: bool, If True, do not refresh the access token even if it
        is out of date. (For use with operations that do not require a current
        access token, such as credential revocation.)
    allow_account_impersonation: bool, True to allow use of impersonated service
      account credentials (if that is configured). If False, the active user
      credentials will always be loaded.
    use_google_auth: bool, True to load credentials as google-auth credentials.
      False to load credentials as oauth2client credentials..
    with_access_token_cache: bool, True to load a credential store with
      auto caching for access tokens. False to load a credential store without
      auto caching for access tokens.

  Returns:
    oauth2client.client.Credentials or google.auth.credentials.Credentials.
    When all of the following conditions are met, it returns
    google.auth.credentials.Credentials and otherwise it returns
    oauth2client.client.Credentials.

    * use_google_auth is True
    * google-auth is not globally disabled by auth/disable_load_google_auth.

  Raises:
    NoActiveAccountException: If account is not provided and there is no
        active account.
    NoCredentialsForAccountException: If there are no valid credentials
        available for the provided or active account.
    c_gce.CannotConnectToMetadataServerException: If the metadata server cannot
        be reached.
    TokenRefreshError: If the credentials fail to refresh.
    TokenRefreshReauthError: If the credentials fail to refresh due to reauth.
    AccountImpersonationError: If impersonation is requested but an
      impersonation provider is not configured.
  """
  use_google_auth = use_google_auth and (not GoogleAuthDisabledGlobally())

  impersonate_service_account = (
      properties.VALUES.auth.impersonate_service_account.Get())
  if allow_account_impersonation and impersonate_service_account:
    target_principal, delegates = ParseImpersonationAccounts(
        impersonate_service_account)
    if not IMPERSONATION_TOKEN_PROVIDER:
      raise AccountImpersonationError(
          'gcloud is configured to impersonate service account [{}] but '
          'impersonation support is not available.'.format(
              target_principal))
    log.warning(
        'This command is using service account impersonation. All API calls '
        'will be executed as [{}].'.format(target_principal))
    if use_google_auth:
      google_auth_source_creds = Load(
          account=account,
          allow_account_impersonation=False,
          use_google_auth=use_google_auth,
          with_access_token_cache=with_access_token_cache)
      cred = IMPERSONATION_TOKEN_PROVIDER.GetElevationAccessTokenGoogleAuth(
          google_auth_source_creds, target_principal, delegates, scopes or
          config.CLOUDSDK_SCOPES)
    else:
      cred = IMPERSONATION_TOKEN_PROVIDER.GetElevationAccessToken(
          impersonate_service_account, scopes or config.CLOUDSDK_SCOPES)
  else:
    cred = _Load(
        account,
        scopes,
        prevent_refresh,
        use_google_auth,
        with_access_token_cache=with_access_token_cache)

  return cred


def _LoadFromFileOverride(cred_file_override, scopes, use_google_auth):
  """Load credentials from cred file override."""
  log.info('Using alternate credentials from file: [%s]', cred_file_override)

  if not use_google_auth:
    try:
      cred = client.GoogleCredentials.from_stream(cred_file_override)
    except client.Error as e:
      raise InvalidCredentialFileException(cred_file_override, e)

    if cred.create_scoped_required():
      if scopes is None:
        scopes = config.CLOUDSDK_SCOPES
      cred = cred.create_scoped(scopes)

    # Set token_uri after scopes since token_uri needs to be explicitly
    # preserved when scopes are applied.
    token_uri_override = properties.VALUES.auth.token_host.Get()
    if token_uri_override:
      cred_type = c_creds.CredentialType.FromCredentials(cred)
      if cred_type in (c_creds.CredentialType.SERVICE_ACCOUNT,
                       c_creds.CredentialType.P12_SERVICE_ACCOUNT):
        cred.token_uri = token_uri_override
    # The credential override is not stored in credential store, but we still
    # want to cache access tokens between invocations.
    cred = c_creds.MaybeAttachAccessTokenCacheStore(cred)
  else:
    google_auth_default = c_creds.GetGoogleAuthDefault()
    # Import only when necessary to decrease the startup time. Move it to
    # global once google-auth is ready to replace oauth2client.
    # Ideally we should wrap the following two lines inside a pylint disable and
    # enable block just like other places, but it seems it is not working here.
    from google.auth import credentials as google_auth_creds  # pylint: disable=g-import-not-at-top

    try:
      # pylint: disable=protected-access
      cred, _ = google_auth_default.load_credentials_from_file(
          cred_file_override)
      # pylint: enable=protected-access
    except google_auth_exceptions.DefaultCredentialsError as e:
      raise InvalidCredentialFileException(cred_file_override, e)

    if scopes is None:
      scopes = config.CLOUDSDK_SCOPES
    cred = google_auth_creds.with_scopes_if_required(cred, scopes)

    if (isinstance(cred, google_auth_external_account.Credentials) and
        not cred.service_account_email):
      # Reinitialize with client auth.
      json_info = cred.info
      json_info['client_id'] = config.CLOUDSDK_CLIENT_ID
      json_info['client_secret'] = config.CLOUDSDK_CLIENT_NOTSOSECRET
      cred = type(cred).from_info(json_info, scopes=config.CLOUDSDK_SCOPES)

    # Set token_uri after scopes since token_uri needs to be explicitly
    # preserved when scopes are applied.
    cred_type = c_creds.CredentialTypeGoogleAuth.FromCredentials(cred)
    if cred_type == c_creds.CredentialTypeGoogleAuth.SERVICE_ACCOUNT:
      token_uri_override = properties.VALUES.auth.token_host.Get()

      if token_uri_override:
        # pylint: disable=protected-access
        cred._token_uri = token_uri_override
        # pylint: enable=protected-access
    elif cred_type == c_creds.CredentialTypeGoogleAuth.USER_ACCOUNT:
      token_uri_override = auth_util.GetTokenUri()
      # pylint: disable=protected-access
      cred._token_uri = token_uri_override
      # pylint: enable=protected-access

    # The credential override is not stored in credential store, but we still
    # want to cache access tokens between invocations.
    cred = c_creds.MaybeAttachAccessTokenCacheStoreGoogleAuth(cred)

  return cred


def _LoadAccessTokenCredsFromValue(access_token, use_google_auth):
  """Loads an AccessTokenCredentials from access_token."""
  log.info('Using access token from environment variable [%s]',
           ACCESS_TOKEN_ENV_VAR_NAME)
  # All commands should be using google-auth, we have this check in case some
  # users disabled google-auth using the hidden auth/disable_load_google_auth.
  if not use_google_auth:
    raise UnsupportedCredentialsError(
        'You may have passed an access token to gcloud using the environment '
        'variable {}. At the same time, google-auth is disabled by '
        'auth/disable_load_google_auth. They do not work together. Please '
        'unset auth/disable_load_google_auth and retry.'.format(
            ACCESS_TOKEN_ENV_VAR_NAME)
    )
  access_token = access_token.strip()
  # Import only when necessary to decrease the startup time. Move it to
  # global once google-auth replaces oauth2client.
  # pylint: disable=g-import-not-at-top
  from googlecloudsdk.core.credentials import google_auth_credentials as c_google_auth
  # pylint: enable=g-import-not-at-top
  return c_google_auth.AccessTokenCredentials(access_token)


def _LoadAccessTokenCredsFromFile(token_file, use_google_auth):
  """Loads an AccessTokenCredentials from token_file."""
  log.info('Using access token from file: [%s]', token_file)
  # All commands should be using google-auth, we have this check in case some
  # users disabled google-auth using the hidden auth/disable_load_google_auth.
  if not use_google_auth:
    raise UnsupportedCredentialsError(
        'You may have passed an access token to gcloud using '
        '--access-token-file or auth/access_token_file. At the same time, '
        'google-auth is disabled by auth/disable_load_google_auth. '
        'They do not work together. Please unset '
        'auth/disable_load_google_auth and retry.'
    )
  content = files.ReadFileContents(token_file).strip()
  # Import only when necessary to decrease the startup time. Move it to
  # global once google-auth replaces oauth2client.
  # pylint: disable=g-import-not-at-top
  from googlecloudsdk.core.credentials import google_auth_credentials as c_google_auth
  # pylint: enable=g-import-not-at-top
  return c_google_auth.AccessTokenCredentials(content)


def _Load(account,
          scopes,
          prevent_refresh,
          use_google_auth=True,
          with_access_token_cache=True):
  """Helper for Load()."""
  # If an access token, access token file, or credential file is set, just use
  # that and ignore the active account and whatever is in the credential store.
  access_token = encoding.GetEncodedValue(os.environ, ACCESS_TOKEN_ENV_VAR_NAME)
  access_token_file = properties.VALUES.auth.access_token_file.Get()
  cred_file_override = properties.VALUES.auth.credential_file_override.Get()

  if access_token:
    cred = _LoadAccessTokenCredsFromValue(access_token, use_google_auth)
  elif access_token_file:
    cred = _LoadAccessTokenCredsFromFile(access_token_file, use_google_auth)
  elif cred_file_override:
    cred = _LoadFromFileOverride(cred_file_override, scopes, use_google_auth)
  else:
    if not account:
      account = properties.VALUES.core.account.Get()

    if not account:
      raise creds_exceptions.NoActiveAccountException(
          named_configs.ActiveConfig(False).file_path)

    cred = STATIC_CREDENTIAL_PROVIDERS.GetCredentials(account, use_google_auth)
    if cred is not None:
      return cred

    store = c_creds.GetCredentialStore(
        with_access_token_cache=with_access_token_cache)
    cred = store.Load(account, use_google_auth)
    if not cred:
      raise NoCredentialsForAccountException(account)

  if not prevent_refresh:
    RefreshIfAlmostExpire(cred)

  return cred


def Refresh(credentials,
            is_impersonated_credential=False,
            include_email=False,
            gce_token_format='standard',
            gce_include_license=False):
  """Refresh credentials.

  Calls credentials.refresh(), unless they're SignedJwtAssertionCredentials.
  If the credentials correspond to a service account or impersonated credentials
  issue an additional request to generate a fresh id_token.

  Args:
    credentials: oauth2client.client.Credentials or
      google.auth.credentials.Credentials, The credentials to refresh.
    is_impersonated_credential: bool, True treat provided credential as an
      impersonated service account credential. If False, treat as service
      account or user credential. Needed to avoid circular dependency on
      IMPERSONATION_TOKEN_PROVIDER.
    include_email: bool, Specifies whether or not the service account email is
      included in the identity token. Only applicable to impersonated service
      account.
    gce_token_format: str, Specifies whether or not the project and instance
      details are included in the identity token. Choices are "standard",
      "full".
    gce_include_license: bool, Specifies whether or not license codes for images
      associated with GCE instance are included in their identity tokens.

  Raises:
    TokenRefreshError: If the credentials fail to refresh.
    TokenRefreshReauthError: If the credentials fail to refresh due to reauth.
  """
  if c_creds.IsOauth2ClientCredentials(credentials):
    _Refresh(credentials, is_impersonated_credential,
             include_email, gce_token_format, gce_include_license)
  else:
    _RefreshGoogleAuth(credentials, is_impersonated_credential,
                       include_email, gce_token_format, gce_include_license)


def _Refresh(credentials,
             is_impersonated_credential=False,
             include_email=False,
             gce_token_format='standard',
             gce_include_license=False):
  """Refreshes oauth2client credentials."""
  http_client = http.Http(response_encoding=transport.ENCODING)
  try:
    credentials.refresh(http_client)
    id_token = None
    # Service accounts require an additional request to receive a fresh id_token
    if is_impersonated_credential:
      if not IMPERSONATION_TOKEN_PROVIDER:
        raise AccountImpersonationError(
            'gcloud is configured to impersonate a service account but '
            'impersonation support is not available.')
      if not IMPERSONATION_TOKEN_PROVIDER.IsImpersonationCredential(
          credentials):
        raise AccountImpersonationError(
            'Invalid impersonation account for refresh {}'.format(credentials))
      id_token = _RefreshImpersonatedAccountIdToken(
          credentials, include_email=include_email)
    # Service accounts require an additional request to receive a fresh id_token
    elif isinstance(credentials, service_account.ServiceAccountCredentials):
      id_token = _RefreshServiceAccountIdToken(credentials, http_client)
    elif isinstance(credentials, oauth2client_gce.AppAssertionCredentials):
      id_token = c_gce.Metadata().GetIdToken(
          config.CLOUDSDK_CLIENT_ID,
          token_format=gce_token_format,
          include_license=gce_include_license)

    if id_token:
      if credentials.token_response:
        credentials.token_response['id_token'] = id_token
      credentials.id_tokenb64 = id_token

  except (client.AccessTokenRefreshError, httplib2.ServerNotFoundError) as e:
    if context_aware.IsContextAwareAccessDeniedError(e):
      raise creds_exceptions.TokenRefreshDeniedByCAAError(e)
    raise creds_exceptions.TokenRefreshError(six.text_type(e))
  except reauth_errors.ReauthSamlLoginRequiredError:
    raise creds_exceptions.WebLoginRequiredReauthError()
  except reauth_errors.ReauthError as e:
    raise creds_exceptions.TokenRefreshReauthError(str(e))


@contextlib.contextmanager
def HandleGoogleAuthCredentialsRefreshError(for_adc=False):
  """Handles exceptions during refreshing google auth credentials."""
  # Import only when necessary to decrease the startup time. Move it to
  # global once google-auth is ready to replace oauth2client.
  # pylint: disable=g-import-not-at-top
  from googlecloudsdk.core.credentials import google_auth_credentials as c_google_auth
  # pylint: enable=g-import-not-at-top
  try:
    yield
  except reauth_errors.ReauthSamlLoginRequiredError:
    raise creds_exceptions.WebLoginRequiredReauthError(for_adc=for_adc)
  except (reauth_errors.ReauthError,
          c_google_auth.ReauthRequiredError) as e:
    raise creds_exceptions.TokenRefreshReauthError(str(e), for_adc=for_adc)
  except google_auth_exceptions.RefreshError as e:
    if context_aware.IsContextAwareAccessDeniedError(e):
      raise creds_exceptions.TokenRefreshDeniedByCAAError(e)
    raise creds_exceptions.TokenRefreshError(six.text_type(e), for_adc=for_adc)


def _RefreshGoogleAuth(credentials,
                       is_impersonated_credential=False,
                       include_email=False,
                       gce_token_format='standard',
                       gce_include_license=False):
  """Refreshes google-auth credentials.

  Args:
    credentials: google.auth.credentials.Credentials, A google-auth credentials
      to refresh.
    is_impersonated_credential: bool, True treat provided credential as an
      impersonated service account credential. If False, treat as service
      account or user credential. Needed to avoid circular dependency on
      IMPERSONATION_TOKEN_PROVIDER.
    include_email: bool, Specifies whether or not the service account email is
      included in the identity token. Only applicable to impersonated service
      account.
    gce_token_format: str, Specifies whether or not the project and instance
      details are included in the identity token. Choices are "standard",
      "full".
    gce_include_license: bool, Specifies whether or not license codes for images
      associated with GCE instance are included in their identity tokens.

  Raises:
    AccountImpersonationError: if impersonation support is not available for
      gcloud, or if the provided credentials is not google auth impersonation
      credentials.
  """
  # Import only when necessary to decrease the startup time. Move it to
  # global once google-auth is ready to replace oauth2client.
  # pylint: disable=g-import-not-at-top
  from google.oauth2 import service_account as google_auth_service_account
  # pylint: enable=g-import-not-at-top
  request_client = requests.GoogleAuthRequest()
  with HandleGoogleAuthCredentialsRefreshError():
    credentials.refresh(request_client)

    id_token = None
    if is_impersonated_credential:
      if not IMPERSONATION_TOKEN_PROVIDER:
        raise AccountImpersonationError(
            'gcloud is configured to impersonate a service account but '
            'impersonation support is not available.')

      # Import only when necessary to decrease the startup time. Move it to
      # global once google-auth is ready to replace oauth2client.
      # pylint: disable=g-import-not-at-top
      import google.auth.impersonated_credentials as google_auth_impersonated_creds
      # pylint: enable=g-import-not-at-top
      if not isinstance(credentials,
                        google_auth_impersonated_creds.Credentials):
        raise AccountImpersonationError(
            'Invalid impersonation account for refresh {}'.format(credentials))
      id_token_creds = IMPERSONATION_TOKEN_PROVIDER.GetElevationIdTokenGoogleAuth(
          credentials, config.CLOUDSDK_CLIENT_ID, include_email)
      id_token_creds.refresh(request_client)
      id_token = id_token_creds.token
    elif isinstance(credentials, google_auth_service_account.Credentials):
      id_token = _RefreshServiceAccountIdTokenGoogleAuth(
          credentials, request_client)
    elif isinstance(credentials, google_auth_gce.Credentials):
      id_token = c_gce.Metadata().GetIdToken(
          config.CLOUDSDK_CLIENT_ID,
          token_format=gce_token_format,
          include_license=gce_include_license)

    # id_token is None (not applicable) for external account credentials.
    if id_token:
      # '_id_token' is the field supported in google-auth natively. gcloud
      # keeps an additional field 'id_tokenb64' to store this information
      # which is referenced in several places
      credentials._id_token = id_token  # pylint: disable=protected-access
      credentials.id_tokenb64 = id_token


def RefreshIfExpireWithinWindow(credentials, window):
  """Refreshes credentials if they will expire within a time window.

  Args:
    credentials: google.auth.credentials.Credentials or
      client.OAuth2Credentials, the credentials to refresh.
    window: string, The threshold of the remaining lifetime of the token which
      can trigger the refresh.
  """
  if c_creds.IsOauth2ClientCredentials(credentials):
    expiry = credentials.token_expiry
  else:
    expiry = credentials.expiry
  almost_expire = (not expiry) or _TokenExpiresWithinWindow(window, expiry)
  if almost_expire:
    Refresh(credentials)


def RefreshIfAlmostExpire(credentials):
  RefreshIfExpireWithinWindow(credentials, window=_CREDENTIALS_EXPIRY_WINDOW)


def _RefreshImpersonatedAccountIdToken(cred, include_email):
  """Get a fresh id_token for the given impersonated service account."""
  # pylint: disable=protected-access
  service_account_email = cred._service_account_id
  return IMPERSONATION_TOKEN_PROVIDER.GetElevationIdToken(
      service_account_email, config.CLOUDSDK_CLIENT_ID, include_email)
  # pylint: enable=protected-access


def _RefreshServiceAccountIdToken(cred, http_client):
  """Get a fresh id_token for the given oauth2client credentials.

  Args:
    cred: service_account.ServiceAccountCredentials, the credentials for which
      to refresh the id_token.
    http_client: httplib2.Http, the http transport to refresh with.

  Returns:
    str, The id_token if refresh was successful. Otherwise None.
  """
  http_request = http_client.request

  now = int(time.time())
  # pylint: disable=protected-access
  payload = {
      'aud': cred.token_uri,
      'iat': now,
      'exp': now + cred.MAX_TOKEN_LIFETIME_SECS,
      'iss': cred._service_account_email,
      'target_audience': config.CLOUDSDK_CLIENT_ID,
  }
  assertion = crypt.make_signed_jwt(
      cred._signer, payload, key_id=cred._private_key_id)

  body = urllib.parse.urlencode({
      'assertion': assertion,
      'grant_type': _GRANT_TYPE,
  })

  resp, content = http_request(
      cred.token_uri.encode('idna'), method='POST', body=body,
      headers=cred._generate_refresh_request_headers())
  # pylint: enable=protected-access
  if resp.status == 200:
    d = json.loads(content)
    return d.get('id_token', None)
  else:
    return None


def _RefreshServiceAccountIdTokenGoogleAuth(cred, request_client):
  """Get a fresh id_token for the given google-auth credentials.

  Args:
    cred: google.oauth2.service_account.Credentials, the credentials for which
      to refresh the id_token.
    request_client: google.auth.transport.Request, the http transport
     to refresh with.

  Returns:
    str, The id_token if refresh was successful. Otherwise None.
  """
  # Import only when necessary to decrease the startup time. Move it to
  # global once google-auth is ready to replace oauth2client.
  # pylint: disable=g-import-not-at-top
  from google.oauth2 import service_account as google_auth_service_account
  # pylint: enable=g-import-not-at-top

  id_token_cred = google_auth_service_account.IDTokenCredentials(
      cred.signer,
      cred.service_account_email,
      cred._token_uri,  # pylint: disable=protected-access
      config.CLOUDSDK_CLIENT_ID)

  try:
    id_token_cred.refresh(request_client)
  except google_auth_exceptions.RefreshError:
    # ID token refresh does not work in testgaia because the Cloud SDK
    # client ID (http://shortn/_BVYwsLLdaJ) is not set up to work with this
    # environment. The running command should not break because of this and
    # should proceed without a new ID token.
    return None

  return id_token_cred.token


def Store(credentials, account=None, scopes=None):
  """Store credentials according for an account address.

  gcloud only stores user account credentials, external account credentials,
  service account credentials and p12 service account credentials. GCE, IAM
  impersonation, and Devshell credentials are generated in runtime.
  External account credentials do not contain any sensitive credentials. They
  only provide hints on how to retrieve local external and exchange them for
  Google access tokens.

  Args:
    credentials: oauth2client.client.Credentials or
      google.auth.credentials.Credentials, The credentials to be stored.
    account: str, The account address of the account they're being stored for.
        If None, the account stored in the core.account property is used.
    scopes: tuple, Custom auth scopes to request. By default CLOUDSDK_SCOPES
        are requested.

  Raises:
    NoActiveAccountException: If account is not provided and there is no
        active account.
  """

  if c_creds.IsOauth2ClientCredentials(credentials):
    cred_type = c_creds.CredentialType.FromCredentials(credentials)
  else:
    cred_type = c_creds.CredentialTypeGoogleAuth.FromCredentials(credentials)

  if cred_type.key not in [
      c_creds.USER_ACCOUNT_CREDS_NAME, c_creds.EXTERNAL_ACCOUNT_CREDS_NAME,
      c_creds.EXTERNAL_ACCOUNT_USER_CREDS_NAME,
      c_creds.SERVICE_ACCOUNT_CREDS_NAME, c_creds.P12_SERVICE_ACCOUNT_CREDS_NAME
  ]:
    return

  if not account:
    account = properties.VALUES.core.account.Get()
  if not account:
    raise creds_exceptions.NoActiveAccountException()

  store = c_creds.GetCredentialStore()
  store.Store(account, credentials)

  _LegacyGenerator(account, credentials, scopes).WriteTemplate()


def ActivateCredentials(account, credentials):
  """Validates, stores and activates credentials with given account."""
  Refresh(credentials)
  Store(credentials, account)

  properties.PersistProperty(properties.VALUES.core.account, account)


def RevokeCredentials(credentials):
  """Revokes the token on the server.

  Args:
    credentials: user account credentials from either google-auth or
      oauth2client.
  Raises:
    RevokeError: If credentials to revoke is not user account credentials.
  """
  if (not c_creds.IsUserAccountCredentials(credentials) or
      # External account user credentials cannot be revoked.
      c_creds.IsExternalAccountUserCredentials(credentials)):
    raise RevokeError('The token cannot be revoked from server because it is '
                      'not user account credentials.')
  if c_creds.IsOauth2ClientCredentials(credentials):
    credentials.revoke(http.Http())
  else:
    credentials.revoke(requests.GoogleAuthRequest())


def Revoke(account=None):
  """Revoke credentials and clean up related files.

  Args:
    account: str, The account address for the credentials to be revoked. If
        None, the currently active account is used.

  Returns:
    True if this call revoked the account; False if the account was already
    revoked.

  Raises:
    NoActiveAccountException: If account is not provided and there is no
        active account.
    NoCredentialsForAccountException: If the provided account is not tied to any
        known credentials.
    RevokeError: If there was a more general problem revoking the account.
  """
  # Import only when necessary to decrease the startup time. Move it to
  # global once google-auth is ready to replace oauth2client.
  # pylint: disable=g-import-not-at-top
  from googlecloudsdk.core.credentials import google_auth_credentials as c_google_auth
  # pylint: enable=g-import-not-at-top
  if not account:
    account = properties.VALUES.core.account.Get()
  if not account:
    raise creds_exceptions.NoActiveAccountException()

  if account in c_gce.Metadata().Accounts():
    raise RevokeError('Cannot revoke GCE-provided credentials.')

  credentials = Load(
      account, prevent_refresh=True, use_google_auth=True)
  if not credentials:
    raise NoCredentialsForAccountException(account)

  rv = False
  try:
    # External account credentials are not revocable.
    if (not account.endswith('.gserviceaccount.com') and
        not isinstance(credentials, google_auth_external_account.Credentials)):
      RevokeCredentials(credentials)
      rv = True
  except (client.TokenRevokeError, c_google_auth.TokenRevokeError) as e:
    if e.args[0] == 'invalid_token':
      # Malformed or already revoked
      pass
    elif e.args[0] == 'invalid_request':
      # Service account token
      pass
    else:
      raise

  store = c_creds.GetCredentialStore()
  store.Remove(account)

  _LegacyGenerator(account, credentials).Clean()
  legacy_creds_dir = config.Paths().LegacyCredentialsDir(account)
  if os.path.isdir(legacy_creds_dir):
    files.RmTree(legacy_creds_dir)
  return rv


def AcquireFromToken(refresh_token,
                     token_uri=None,
                     revoke_uri=GOOGLE_OAUTH2_PROVIDER_REVOKE_URI,
                     use_google_auth=True):
  """Get credentials from an already-valid refresh token.

  Args:
    refresh_token: An oauth2 refresh token.
    token_uri: str, URI to use for refreshing.
    revoke_uri: str, URI to use for revoking.
    use_google_auth: bool, True to return google-auth credentials. False to
    return oauth2client credentials..

  Returns:
    oauth2client.client.Credentials or google.auth.credentials.Credentials.
    When all of the following conditions are true, it returns
    google.auth.credentials.Credentials and otherwise it returns
    oauth2client.client.Credentials.

    * use_google_auth=True
    * google-auth is not globally disabled by auth/disable_load_google_auth.
  """
  use_google_auth = use_google_auth and (not GoogleAuthDisabledGlobally())

  if token_uri is None:
    token_uri = auth_util.GetTokenUri()
  if use_google_auth:
    # Import only when necessary to decrease the startup time. Move it to
    # global once google-auth is ready to replace oauth2client.
    # pylint: disable=g-import-not-at-top
    from google.oauth2 import credentials as google_auth_creds
    # pylint: enable=g-import-not-at-top
    cred = google_auth_creds.Credentials(
        token=None,
        refresh_token=refresh_token,
        id_token=None,
        token_uri=token_uri,
        client_id=properties.VALUES.auth.client_id.Get(required=True),
        client_secret=properties.VALUES.auth.client_secret.Get(required=True))

    # always start expired
    cred.expiry = datetime.datetime.utcnow()
  else:
    cred = client.OAuth2Credentials(
        access_token=None,
        client_id=properties.VALUES.auth.client_id.Get(required=True),
        client_secret=properties.VALUES.auth.client_secret.Get(required=True),
        refresh_token=refresh_token,
        # always start expired
        token_expiry=datetime.datetime.utcnow(),
        token_uri=token_uri,
        user_agent=config.CLOUDSDK_USER_AGENT,
        revoke_uri=revoke_uri)
  return cred


def AcquireFromGCE(account=None, use_google_auth=True):
  """Get credentials from a GCE metadata server.

  Args:
    account: str, The account name to use. If none, the default is used.
    use_google_auth: bool, True to load credentials of google-auth if it is
      supported in the current authentication scenario. False to load
      credentials of oauth2client.

  Returns:
    oauth2client.client.Credentials or google.auth.credentials.Credentials based
    on use_google_auth and whether google-auth is supported in the current
    authentication sceanrio.

  Raises:
    c_gce.CannotConnectToMetadataServerException: If the metadata server cannot
      be reached.
    TokenRefreshError: If the credentials fail to refresh.
    TokenRefreshReauthError: If the credentials fail to refresh due to reauth.
  """
  if use_google_auth:
    email = account or 'default'
    credentials = google_auth_gce.Credentials(service_account_email=email)
  else:
    credentials = oauth2client_gce.AppAssertionCredentials(email=account)
  Refresh(credentials)
  return credentials


class _LegacyGenerator(object):
  """A class to generate the credential file for other tools, like gsutil & bq.

  The supported credentials types are user account credentials, service account
  credentials, and p12 service account credentials. Gcloud supports two auth
  libraries - oauth2client and google-auth. Eventually, we will deprecate
  oauth2client.
  """

  def __init__(self, account, credentials, scopes=None):
    self.credentials = credentials
    if self._cred_type not in (c_creds.USER_ACCOUNT_CREDS_NAME,
                               c_creds.SERVICE_ACCOUNT_CREDS_NAME,
                               c_creds.EXTERNAL_ACCOUNT_CREDS_NAME,
                               c_creds.EXTERNAL_ACCOUNT_USER_CREDS_NAME,
                               c_creds.P12_SERVICE_ACCOUNT_CREDS_NAME):
      raise c_creds.CredentialFileSaveError(
          'Unsupported credentials type {0}'.format(type(self.credentials)))
    if scopes is None:
      self.scopes = config.CLOUDSDK_SCOPES
    else:
      self.scopes = scopes

    paths = config.Paths()
    # Bq file is not generated here. bq CLI generates it using the adc at
    # self._adc_path and uses it as the cache.
    # Register so it is cleaned up.
    self._bq_path = paths.LegacyCredentialsBqPath(account)
    self._gsutil_path = paths.LegacyCredentialsGSUtilPath(account)
    self._p12_key_path = paths.LegacyCredentialsP12KeyPath(account)
    self._adc_path = paths.LegacyCredentialsAdcPath(account)

  @property
  def _is_oauth2client(self):
    return c_creds.IsOauth2ClientCredentials(self.credentials)

  @property
  def _cred_type(self):
    if self._is_oauth2client:
      return c_creds.CredentialType.FromCredentials(self.credentials).key
    else:
      return c_creds.CredentialTypeGoogleAuth.FromCredentials(
          self.credentials).key

  def Clean(self):
    """Remove the credential file."""

    paths = [
        self._bq_path,
        self._gsutil_path,
        self._p12_key_path,
        self._adc_path,
    ]
    for p in paths:
      try:
        os.remove(p)
      except OSError:
        # file did not exist, so we're already done.
        pass

  def WriteTemplate(self):
    """Write the credential file."""

    # Remove all the credential files first. As per the comment in __init__
    # the BQ file (singlestore_bq) is created by the BQ CLI and not
    # regenerated. This file should be removed when the credentials are
    # created. If this file isn't removed, it will be stale, but BQ CLI will
    # try to use it anyways. The rest of the credential files should be
    # recreated here.
    self.Clean()

    # Generates credentials used by bq and gsutil.
    if self._cred_type == c_creds.P12_SERVICE_ACCOUNT_CREDS_NAME:
      cred = self.credentials
      key = cred._private_key_pkcs12  # pylint: disable=protected-access
      password = cred._private_key_password  # pylint: disable=protected-access
      files.WriteBinaryFileContents(self._p12_key_path, key, private=True)

      # the .boto file gets some different fields
      self._WriteFileContents(
          self._gsutil_path, '\n'.join([
              '[Credentials]',
              'gs_service_client_id = {account}',
              'gs_service_key_file = {key_file}',
              'gs_service_key_file_password = {key_password}',
          ]).format(account=self.credentials.service_account_email,
                    key_file=self._p12_key_path,
                    key_password=password))
      return

    c_creds.ADC(
        self.credentials).DumpADCToFile(file_path=self._adc_path)

    if (self._cred_type == c_creds.EXTERNAL_ACCOUNT_CREDS_NAME or
        self._cred_type == c_creds.EXTERNAL_ACCOUNT_USER_CREDS_NAME):
      self._WriteFileContents(
          self._gsutil_path, '\n'.join([
              '[Credentials]',
              'gs_external_account_file = {external_account_file}',
          ]).format(external_account_file=self._adc_path))
      return

    if self._cred_type == c_creds.USER_ACCOUNT_CREDS_NAME:
      # We create a small .boto file for gsutil, to be put in BOTO_PATH.
      # Our client_id and client_secret should accompany our refresh token;
      # if a user loaded any other .boto files that specified a different
      # id and secret, those would override our id and secret, causing any
      # attempts to obtain an access token with our refresh token to fail.
      self._WriteFileContents(
          self._gsutil_path, '\n'.join([
              '[OAuth2]',
              'client_id = {cid}',
              'client_secret = {secret}',
              '',
              '[Credentials]',
              'gs_oauth2_refresh_token = {token}',
          ]).format(cid=config.CLOUDSDK_CLIENT_ID,
                    secret=config.CLOUDSDK_CLIENT_NOTSOSECRET,
                    token=self.credentials.refresh_token))
    else:
      self._WriteFileContents(
          self._gsutil_path, '\n'.join([
              '[Credentials]',
              'gs_service_key_file = {key_file}',
          ]).format(key_file=self._adc_path))

  def _WriteFileContents(self, filepath, contents):
    """Writes contents to a path, ensuring mkdirs.

    Args:
      filepath: str, The path of the file to write.
      contents: str, The contents to write to the file.
    """

    full_path = os.path.realpath(files.ExpandHomeDir(filepath))
    files.WriteFileContents(full_path, contents, private=True)



# -*- coding: utf-8 -*- #
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers for dealing with text."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from dateutil import tz
from googlecloudsdk.core.util import times


def SnakeCaseToCamelCase(name):
  words = name.split('_')
  return words[0].lower() + ''.join(
      [w[0].upper() + w[1:].lower() for w in words[1:]])


def ToSnakeCaseDict(dictionary):
  """Recursively convert all keys in nested dictionaries to snakeCase."""
  new_dict = {}
  for key, val in dictionary.items():
    snaked_key = SnakeCaseToCamelCase(key)
    if isinstance(val, dict):
      new_dict[snaked_key] = ToSnakeCaseDict(val)
    else:
      new_dict[snaked_key] = val

  return new_dict


def TransformNotBeforeTime(subject_description):
  """Use this function in a display transform to truncate anything smaller than minutes from ISO8601 timestamp."""
  if subject_description and 'notBeforeTime' in subject_description:
    return times.ParseDateTime(
        subject_description.get('notBeforeTime')).astimezone(
            tz.tzutc()).strftime('%Y-%m-%dT%H:%MZ')
  return ''


def TransformNotAfterTime(subject_description):
  """Use this function in a display transform to truncate anything smaller than minutes from ISO8601 timestamp."""
  if subject_description and 'notAfterTime' in subject_description:
    return times.ParseDateTime(
        subject_description.get('notAfterTime')).astimezone(
            tz.tzutc()).strftime('%Y-%m-%dT%H:%MZ')
  return ''



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of hash command for getting formatted file hashes."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.calliope import base


@base.Hidden
@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class Hash(base.Command):
  """Calculates hashes on local or cloud files."""

  detailed_help = {
      'DESCRIPTION':
          """
      Calculates hashes on local or cloud files that can be used to compare with
      "gcloud storage ls -L" output. If a specific hash option is not provided,
      this command calculates all gcloud storage-supported hashes for the file.

      Note that gcloud storage automatically performs hash validation when
      uploading or downloading files, so this command is only needed if you want
      to write a script that separately checks the hash for some reason.

      If you calculate a CRC32C hash for the file without a precompiled
      google-crc32c installation, hashing will be very slow.
      """,
      'EXAMPLES':
          """

      To get the MD5 hash of a cloud object:

        $ {command} gs://bucket/object --md5

      To get the CRC32C hash of a local object in hexadecimal:

        $ {command} /dir/object.txt --crc32c --hex
      """,
  }

  @staticmethod
  def Args(parser):
    parser.add_argument(
        'url', nargs='+', help='Local or cloud URLs of objects to hash.')
    parser.add_argument(
        '-c',
        '--crc32c',
        action='store_true',
        help='Calculate a CRC32c hash for the file.')
    parser.add_argument(
        '--hex',
        action='store_true',
        help='Output hashes in hex format. By default, gsutil uses base64.')
    parser.add_argument(
        '-m',
        '--md5',
        action='store_true',
        help='Calculate a MD5 hash for the file.')

  def Run(self, args):
    del args  # Unused.
    raise NotImplementedError



# -*- coding: utf-8 -*- #
# Copyright 2018 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes to specify concept and resource specs.

To use a concept, give it at least help text and a name (or use
the default name if the concept provides one) and add it to a concept manager.
During command.Run, the parsed concept will be available under args.
For example:

from googlecloudsdk.command_lib.concepts import concept_managers

  def Args(self, parser):
    manager = concept_managers.ConceptManager()
    concept = concepts.SimpleArg('foo', help_text='Provide the value of foo.')
    manager.AddConcept(concept)
    manager.AddToParser(parser)

  def Run(self, args):
    return args.foo
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import copy
import io
import re

from googlecloudsdk.calliope.concepts import deps as deps_lib
from googlecloudsdk.command_lib.concepts import base
from googlecloudsdk.command_lib.concepts import dependency_managers
from googlecloudsdk.command_lib.concepts import exceptions
from googlecloudsdk.command_lib.concepts import names
from googlecloudsdk.core.util import scaled_integer
from googlecloudsdk.core.util import semver
from googlecloudsdk.core.util import times

import six


def _SubException(e):
  """Return the string representation of sub-exception e."""
  message = six.text_type(e).rstrip('.')
  return message[0].upper() + message[1:]


def _Insert(text):
  """Appends a space to text if it is not empty and returns it."""
  if not text:
    return ''
  if text[-1].isspace():
    return text
  return text + ' '


def _Append(text):
  """Inserts a space to text if it is not empty and returns it."""
  if not text:
    return ''
  if text[-1].isspace():
    return text
  return ' ' + text


class SimpleArg(base.Concept):
  """A basic concept with a single attribute.

  Attributes:
    fallthroughs: [calliope.concepts.deps.Fallthrough], the list of sources of
      data, in priority order, that can provide a value for the attribute if
      not given on the command line. These should only be sources inherent to
      the attribute, such as associated properties, not command- specific
      sources.
    positional: bool, True if the concept is a positional value.
    completer: core.cache.completion_cache.Completer, the completer associated
      with the attribute.
    metavar: string,  a name for the argument in usage messages.
    default: object, the concept value if one is not otherwise specified.
    choices: {name: help}, the possible concept values with help text.
    action: string or argparse.Action, the basic action to take when the
       concept is specified on the command line. Required for the current
       underlying argparse implementation.
  """

  def __init__(self, name, fallthroughs=None, positional=False, completer=None,
               metavar=None, default=None, choices=None, action=None, **kwargs):
    """Initializes the concept."""
    if name is None:
      raise exceptions.InitializationError('Concept name required.')
    self.fallthroughs = fallthroughs or []
    self.positional = positional
    self.completer = completer
    self.metavar = metavar
    self.default = default
    self.choices = choices
    self.action = action
    super(SimpleArg, self).__init__(name, **kwargs)

  def Attribute(self):
    return base.Attribute(concept=self,
                          fallthroughs=self.fallthroughs,
                          completer=self.completer,
                          metavar=self.metavar,
                          default=self.default,
                          action=self.action,
                          choices=self.choices,
                          **self.MakeArgKwargs())

  def Constraints(self):
    """Returns the type constraints message text if any.

    This message text decribes the Validate() method constraints in English.
    For example, a regex validator could provide prose for a better UX than
    a raw 100 char regex pattern.
    """
    return ''

  def Parse(self, dependencies):
    """Parses the concept.

    Args:
      dependencies: googlecloudsdk.command_lib.concepts.dependency_managers
        .DependencyView, the dependency namespace for the concept.

    Raises:
      exceptions.MissingRequiredArgumentException, if no value is provided and
        one is required.

    Returns:
      str, the value given to the argument.
    """
    try:
      return dependencies.value
    except deps_lib.AttributeNotFoundError as e:
      if self.required:
        raise exceptions.MissingRequiredArgumentError(
            self.GetPresentationName(), _SubException(e))
      return None

  def GetPresentationName(self):
    """Gets presentation name for the attribute, either positional or flag."""
    if self.positional:
      return names.ConvertToPositionalName(self.name)
    return names.ConvertToFlagName(self.name)

  def IsArgRequired(self):
    """Determines whether command line argument for attribute is required.

    Returns:
      bool: True, if the command line argument is required to be provided,
        meaning that the attribute is required and that there are no
        fallthroughs. There may still be a parsing error if the argument isn't
        provided and none of the fallthroughs work.
    """
    return self.required and not self.fallthroughs


class GroupArg(base.Concept):
  """A group concept.

  Attributes:
    mutex: bool, True if this is a mutex (mutually exclusive) group.
  """

  def __init__(self, name, mutex=False, prefixes=False, **kwargs):
    """Initializes the concept."""
    if name is None:
      raise exceptions.InitializationError('Concept name required.')
    self.mutex = mutex
    self.prefixes = prefixes
    self.concepts = []
    super(GroupArg, self).__init__(name, **kwargs)

  def AddConcept(self, concept):
    new_concept = copy.copy(concept)
    new_concept.name = self._GetSubConceptName(new_concept.name)
    self.concepts.append(new_concept)

  def Attribute(self):
    return base.AttributeGroup(
        concept=self,
        attributes=[c.Attribute() for c in self.concepts],
        mutex=self.mutex,
        **self.MakeArgKwargs()
    )

  def _GetSubConceptName(self, attribute_name):
    if self.prefixes:
      return names.ConvertToNamespaceName(self.name + '_' + attribute_name)
    return attribute_name

  def Parse(self, dependencies):
    """Returns a namespace with the values of the child concepts."""
    return dependencies

  def GetPresentationName(self):
    """Gets presentation name for the attribute group."""
    return self.name

  def IsArgRequired(self):
    """Determines whether the concept group is required to be specified.

    Returns:
      bool: True, if the command line argument is required to be provided,
        meaning that the attribute is required and that there are no
        fallthroughs. There may still be a parsing error if the argument isn't
        provided and none of the fallthroughs work.
    """
    return self.required and not any(c.fallthroughs for c in self.concepts)


class ConceptType(SimpleArg):
  """Concept type base class.

  All concept types derive from this class. The methods implement lexing,
  parsing, constraints, help text, and formatting.
  """

  def Convert(self, string):
    """Converts a value from string and returns it.

    The converter must do syntax checking and raise actionable exceptions. All
    non-space characters in string must be consumed. This method may raise
    syntax exceptions, but otherwise does no validation.

    Args:
        string: The string to convert to a concept type value.

    Returns:
      The converted value.
    """
    return string

  def Display(self, value):
    """Returns the string representation of a parsed concept value.

    This method is the inverse of Convert() and Parse(). It returns the
    string representation of a parsed concept value that can be used in
    formatted messages.

    Args:
        value: The concept value to display.

    Returns:
      The string representation of a parsed concept value.
    """
    return six.text_type(value)

  def Normalize(self, value):
    """Returns the normalized value.

    Called after the value has been validated. It normalizes internal values
    for compatibility with other interfaces. This can be accomplished by
    subclassing with a shim class that contains only a Normalize() method.

    Args:
        value: The concept value to normalize.

    Returns:
      The normalized value.
    """
    return value

  def Parse(self, dependencies):
    """Converts, validates and normalizes a value string from dependencies."""
    string = super(ConceptType, self).Parse(dependencies)
    value = self.Convert(string)
    self.Validate(value)
    return self.Normalize(value)

  def Validate(self, value):
    """Validates value.

    Syntax checking has already been done by Convert(). The validator imposes
    additional constraints on valid values for the concept type and must raise
    actionable exceptions when the constraints are not met.

    Args:
      value: The concept value to validate.
    """
    pass


class Endpoint(object):
  """TypeWithIntervalConstraint endpoint.

  Attributes:
    string: string, the representation of the endpoint value.
    closed: bool, True if the interval is closed (the endpoint is included).
  """

  def __init__(self, string, closed=True):
    self.string = string
    self.closed = closed
    self.value = None


class TypeWithIntervalConstraint(ConceptType):
  """Concept type with value interval constraints.

  Validates that a ConceptType value is within the interval defined by min and
  max endpoints. A missing min or max endpoint indicates that there is no min or
  max value, respectively.

  Attributes:
    _min_endpoint: Endpoint, the minimum value interval endpoint.
    _max_endpoint: Endpoint, the maximum value interval endpoint.
    _constraint_kind: string, the interval value type name.
    _convert_endpoint: f(str)=>x, converts an endpoint string to a value.
    _convert_interval: f(str)=>x, converts an interval value to a value.
    _display_endpoint: f(value)=>str, displays an interval endpoint.
  """

  def __init__(self, name, min_endpoint=None, max_endpoint=None,
               constraint_kind=None, convert_endpoint=None, convert_value=None,
               display_endpoint=None, **kwargs):
    super(TypeWithIntervalConstraint, self).__init__(name, **kwargs)
    self._min_endpoint = min_endpoint
    self._max_endpoint = max_endpoint
    self._kind = constraint_kind or 'value'
    self._convert_endpoint = convert_endpoint or self.Convert
    self._display_endpoint = display_endpoint or self.Display
    self._convert_value = convert_value or (lambda x: x)
    if self._min_endpoint:
      self._ConvertEndpoint(self._min_endpoint, 'min endpoint')
    if self._max_endpoint:
      self._ConvertEndpoint(self._max_endpoint, 'max endpoint')

  def _ConvertEndpoint(self, endpoint, kind):
    """Declaration time endpoint conversion check."""
    message = None
    try:
      endpoint.value = self._convert_endpoint(endpoint.string)
      return
    except exceptions.ParseError as e:
      message = six.text_type(e).split('. ', 1)[1].rstrip('.')
    except (AttributeError, ValueError) as e:
      message = _SubException(e)
    raise exceptions.ConstraintError(
        self.GetPresentationName(), kind, endpoint.string, message + '.')

  def Constraints(self):
    """Returns the type constraints message text if any."""
    boundaries = []
    if self._min_endpoint:
      endpoint = self._min_endpoint.value
      if self._min_endpoint.closed:
        boundary = 'greater than or equal to'
      else:
        boundary = 'greater than'
      boundaries.append('{} {}'.format(
          boundary, self._display_endpoint(endpoint)))
    if self._max_endpoint:
      endpoint = self._max_endpoint.value
      if self._max_endpoint.closed:
        boundary = 'less than or equal to'
      else:
        boundary = 'less than'
      boundaries.append('{} {}'.format(
          boundary, self._display_endpoint(endpoint)))
    if not boundaries:
      return ''
    return 'The {} must be {}.'.format(self._kind, ' and '.join(boundaries))

  def Validate(self, value):
    value = self._convert_value(value)
    invalid = None
    if self._min_endpoint:
      endpoint = self._min_endpoint.value
      if self._min_endpoint.closed:
        if value < endpoint:
          invalid = 'greater than or equal to'
      elif value <= endpoint:
        invalid = 'greater than'
    if not invalid and self._max_endpoint:
      endpoint = self._max_endpoint.value
      if self._max_endpoint.closed:
        if value > endpoint:
          invalid = 'less than or equal to'
      elif value >= endpoint:
        invalid = 'less than'
    if invalid:
      raise exceptions.ValidationError(
          self.GetPresentationName(),
          '{}{} [{}] must be {} [{}].'.format(
              self._kind[0].upper(),
              self._kind[1:],
              self._display_endpoint(value),
              invalid,
              self._display_endpoint(endpoint)))


class TypeWithSizeConstraint(TypeWithIntervalConstraint):
  """Concept type with size interval constraints.

  Validates that a ConceptType size is within the interval defined by min and
  max endpoints. A missing min or max endpoint indicates that there is no min or
  max size, respectively.
  """

  _DEFAULT_DELIM = ','
  _ALT_DELIM = '^'

  @classmethod
  def _GetIntervalValue(cls, value):
    return len(value) if value else 0

  def __init__(self, name, constraint_kind=None, convert_endpoint=None,
               convert_value=None, display_endpoint=None, **kwargs):
    super(TypeWithSizeConstraint, self).__init__(
        name,
        constraint_kind=constraint_kind or 'size',
        convert_endpoint=convert_endpoint or int,
        convert_value=convert_value or self._GetIntervalValue,
        display_endpoint=convert_endpoint or str,
        **kwargs)

  def _Split(self, string):
    """Splits string on _DEFAULT_DELIM or the alternate delimiter expression.

    By default, splits on commas:
        'a,b,c' -> ['a', 'b', 'c']

    Alternate delimiter syntax:
        '^:^a,b:c' -> ['a,b', 'c']
        '^::^a:b::c' -> ['a:b', 'c']
        '^,^^a^,b,c' -> ['^a^', ',b', 'c']

    See `gcloud topic escaping` for details.

    Args:
      string: The string with optional alternate delimiter expression.

    Raises:
      exceptions.ParseError: on invalid delimiter expression.

    Returns:
      (string, delimiter) string with the delimiter expression stripped, if any.
    """
    if not string:
      return None, None
    delim = self._DEFAULT_DELIM
    if string.startswith(self._ALT_DELIM) and self._ALT_DELIM in string[1:]:
      delim, string = string[1:].split(self._ALT_DELIM, 1)
      if not delim:
        raise exceptions.ParseError(
            self.GetPresentationName(),
            'Invalid delimiter. Please see $ gcloud topic escaping for '
            'information on escaping list or dictionary flag values.')
    return string.split(delim), delim


class TypeWithRegexConstraint(ConceptType):
  """Concept type with regex constraint.

  Attributes:
    _regex: string, an unanchored regular expression pattern that must match
      valid values.
    _constraint_details: string, optional prose that describes the regex
      constraint.
  """

  def __init__(self, name, regex=None, constraint_details=None, **kwargs):
    super(TypeWithRegexConstraint, self).__init__(name, **kwargs)
    self._regex = regex
    self._constraint_details = constraint_details

  def Constraints(self):
    """Returns the type constraints message text if any."""
    if not self._regex:
      return ''
    if self._constraint_details:
      return self._constraint_details
    return 'The value must match the regular expression ```{}```.'.format(
        self._regex)

  def Validate(self, value):
    if self._regex and not re.match(self._regex, self.Display(value)):
      raise exceptions.ValidationError(
          self.GetPresentationName(),
          'Value [{}] does not match [{}].'.format(
              self.Display(value),
              self._regex))


class Boolean(ConceptType):
  """Boolean value concept."""

  def __init__(self, name, default=None, **kwargs):
    action = 'store_false' if default else 'store_true'
    help_text = kwargs.get('help_text', 'A Boolean value.')
    if default:
      presentation = self.GetPresentationName()
      if presentation.startswith('--'):
        help_text += ' On by default, use --no-{} to disable.'.format(
            presentation[2:])
    kwargs['help_text'] = help_text
    super(Boolean, self).__init__(name, default=default, action=action,
                                  **kwargs)

  def Convert(self, string):
    # TODO(b/117144623): add and use an IsSpecified() method.
    if string is None:
      return False
    if string == '1' or string.lower() == 'true':
      return True
    if string == '0' or string.lower() == 'false':
      return False
    raise exceptions.ParseError(
        self.GetPresentationName(),
        'Invalid Boolean value [{}].'.format(string))

  def Display(self, value):
    """Returns the display string for a Boolean value."""
    return 'true' if value else 'false'


class Enum(ConceptType):
  """Enum value concept."""

  def __init__(self, name, choices=None, default=None, **kwargs):
    if not choices:
      raise exceptions.InitializationError(
          'Choices must be specified for Enum type.')
    if default and default not in choices:
      raise exceptions.InitializationError(
          'Enum default value must be a valid choice.')
    self.choices = choices
    super(Enum, self).__init__(name, choices=choices, default=default, **kwargs)

  def Convert(self, string):
    try:
      choice = string.lower()
      if choice not in self.choices and choice.upper() not in self.choices:
        raise exceptions.ParseError(
            self.GetPresentationName(),
            'Invalid choice [{}], must be one of [{}].'.format(
                string, ','.join(sorted(self.choices.keys()))))
      return choice
    except (AttributeError, ValueError):
      return string

  def BuildHelpText(self):
    """Appends enum values to the original help text."""
    buf = io.StringIO()
    buf.write('\n+\n')
    for key, help_text in sorted(six.iteritems(self.choices)):
      buf.write('*{}*::: {}\n'.format(key, help_text))
    return (
        '{}Must be one of the following values:{}'.format(
            _Insert(super(Enum, self).BuildHelpText()), buf.getvalue()))


class String(TypeWithRegexConstraint):
  """String value concept."""

  def __init__(self, name, **kwargs):
    if 'help_text' not in kwargs:
      kwargs['help_text'] = 'A string value.'
    super(String, self).__init__(name, **kwargs)


class Integer(TypeWithIntervalConstraint):
  """Integer value concept.

  Attributes:
    _unlimited: bool, the value 'unlimited' specifies the largest valid value.
      Internally it's represented as None.
  """

  def __init__(self, name, unlimited=False, **kwargs):
    self._unlimited = unlimited
    super(Integer, self).__init__(name, **kwargs)

  def BuildHelpText(self):
    """Appends integer syntax to the original help text."""
    return (
        '{}Must be a string representing an integer.'.format(
            _Insert(super(Integer, self).BuildHelpText())))

  def Convert(self, string):
    if string is None:
      return None
    try:
      return int(string)
    except ValueError as e:
      if self._unlimited and string == 'unlimited':
        return None
      raise exceptions.ParseError(
          self.GetPresentationName(),
          '{}.'.format(_SubException(e)))


class ScaledInteger(TypeWithIntervalConstraint):
  """ISO Decimal/Binary scaled Integer value concept.

  ISO/IEC prefixes: 1k == 1000, 1ki == 1024.

  Attributes:
    _default_unit: string, the unit suffix if none is specified.
    _output_unit: string, the implicit output unit. Integer values are
      divided by the output unit value.
    _output_unit_value: int, the output unit value.
    _type_abbr: string, the type abbreviation, for example 'b/s' or 'Hz'.
    _type_details: string, prose that describes type syntax details.
  """

  def __init__(self, name, default_unit=None, output_unit=None, type_abbr='B',
               type_details=None, **kwargs):
    self.name = name
    self._type_abbr = type_abbr
    self._default_unit = default_unit
    if self._default_unit:
      self._default_unit, _ = self._GetUnitValue(
          'default scaled integer unit', self._default_unit)
    self._output_unit = output_unit
    if self._output_unit:
      self._output_unit, self._output_unit_value = self._GetUnitValue(
          'output scaled integer unit', self._output_unit)
    else:
      self._output_unit_value = 0
    self._type_details = type_details or (
        'Must be a string representing an ISO/IEC Decimal/Binary scaled '
        'integer. For example, 1k == 1000 and 1ki == 1024. ')
    super(ScaledInteger, self).__init__(name, **kwargs)

  def _GetUnitValue(self, kind, unit):
    """Returns the integer unit suffix and value for unit."""
    if self._type_abbr:
      unit = scaled_integer.DeleteTypeAbbr(unit)
    try:
      return unit, scaled_integer.GetUnitSize(unit)
    except ValueError as e:
      raise exceptions.ConstraintError(
          self.name, kind, unit, _SubException(e) + '.')

  def BuildHelpText(self):
    """Appends ISO Decimal/Binary scaled integer syntax to the help text."""
    if self._default_unit:
      default_unit = 'The default unit is `{}`. '.format(self._default_unit)
    else:
      default_unit = ''
    if self._output_unit:
      output_unit = (
          'The output unit is `{}`. Integer values are divided by the unit '
          'value. '.format(self._output_unit))
    else:
      output_unit = ''
    if self._type_abbr:
      type_abbr = 'The default type abbreviation is `{}`. '.format(
          self._type_abbr)
    else:
      type_abbr = ''
    return (
        '{}{}{}{}{}{}See https://en.wikipedia.org/wiki/Binary_prefix for '
        'details.'.format(
            _Insert(super(ScaledInteger, self).BuildHelpText()),
            self._type_details,
            default_unit,
            output_unit,
            type_abbr,
            _Insert(self.Constraints())))

  def Convert(self, string):
    if not string:
      return None
    try:
      value = scaled_integer.ParseInteger(
          string, default_unit=self._default_unit, type_abbr=self._type_abbr)
      if self._output_unit_value:
        value //= self._output_unit_value
      return value
    except ValueError as e:
      raise exceptions.ParseError(
          self.GetPresentationName(),
          'Failed to parse binary/decimal scaled integer [{}]: {}.'.format(
              string, _SubException(e)))

  def Display(self, value):
    """Returns the display string for a binary scaled value."""
    if self._output_unit_value:
      value *= self._output_unit_value
    return scaled_integer.FormatInteger(value, type_abbr=self._type_abbr)


class BinaryScaledInteger(ScaledInteger):
  """ISO Binary scaled Integer value with binary display concept.

  All ISO/IEC prefixes are powers of 2: 1k == 1ki == 1024. This is a
  concession to the inconsistent mix of binary/decimal scaled measures for
  memory capacity, disk capacity, cpu speed. Ideally ScaledInteger should
  be used.
  """

  def __init__(self, name, type_details=None, **kwargs):
    if type_details is None:
      type_details = (
          'Must be a string representing binary scaled integer where all '
          'ISO/IEC prefixes are powers of 2. For example, 1k == 1ki == 1024. ')
    super(BinaryScaledInteger, self).__init__(
        name, type_details=type_details, **kwargs)

  def Convert(self, string):
    if not string:
      return None
    try:
      return scaled_integer.ParseBinaryInteger(string, self._default_unit)
    except ValueError as e:
      raise exceptions.ParseError(
          self.GetPresentationName(),
          'Failed to parse binary scaled integer [{}]: {}.'.format(
              string, _SubException(e)))


class Float(TypeWithIntervalConstraint):
  """Float value concept.

  Attributes:
    _unlimited: bool, the value 'unlimited' specifies the largest valid value.
      Internally it's represented as None.
  """

  def __init__(self, name, unlimited=False, **kwargs):
    self._unlimited = unlimited
    super(Float, self).__init__(name, **kwargs)

  def BuildHelpText(self):
    """Appends float syntax to the original help text."""
    return (
        '{}Must be a string representing a floating point number.{}'.format(
            _Insert(super(Float, self).BuildHelpText()),
            _Append(self.Constraints())))

  def Convert(self, string):
    if not string:
      return None
    try:
      return float(string)
    except ValueError as e:
      if self._unlimited and string == 'unlimited':
        return None
      raise exceptions.ParseError(
          self.GetPresentationName(),
          'Failed to parse floating point number [{}]: {}.'.format(
              string, _SubException(e)))


class DayOfWeek(TypeWithRegexConstraint):
  """Day of the week concept."""

  _DAYS = ['SUN', 'MON', 'TUE', 'WED', 'THU', 'FRI', 'SAT']

  def BuildHelpText(self):
    """Appends day of week syntax to the original help text."""
    return (
        '{}Must be a string representing a day of the week in English, '
        'such as \'MON\' or \'FRI\'. Case is ignored, and any characters after '
        'the first three are ignored.{}'.format(
            _Insert(super(DayOfWeek, self).BuildHelpText()),
            _Append(self.Constraints())))

  def Convert(self, string):
    """Converts a day of week from string returns it."""
    if not string:
      return None
    value = string.upper()[:3]
    if value not in self._DAYS:
      raise exceptions.ParseError(
          self.GetPresentationName(),
          'A day of week value [{}] must be one of: [{}].'.format(
              string, ', '.join(self._DAYS)))
    return value


class Duration(TypeWithIntervalConstraint):
  """Duration concept.

  Attributes:
    _default_suffix: string, the time unit suffix if none is specified.
    _subsecond: bool, return floating point values if True.
  """

  def __init__(self, name, default_suffix='s', subsecond=False, **kwargs):
    self._default_suffix = default_suffix
    self._subsecond = subsecond
    super(Duration, self).__init__(name, **kwargs)

  def BuildHelpText(self):
    """Appends duration syntax to the original help text."""
    if self._default_suffix:
      default_suffix = 'The default suffix is `{}`. '.format(
          self._default_suffix)
    else:
      default_suffix = ''
    return (
        '{}Must be a string representing an ISO 8601 duration. Syntax is '
        'relaxed to ignore case, the leading P and date/time separator T if '
        'there is no ambiguity. {}For example, `PT1H` and `1h` are equivalent '
        'representations of one hour. {}See $ gcloud topic datetimes for more '
        'information.'.format(
            _Insert(super(Duration, self).BuildHelpText()), default_suffix,
            _Insert(self.Constraints())))

  def Convert(self, string):
    """Converts a duration from string returns it."""
    if not string:
      return None
    try:
      d = times.ParseDuration(
          string, default_suffix=self._default_suffix).total_seconds
      return d if self._subsecond else int(d)
    except times.Error as e:
      raise exceptions.ParseError(
          self.GetPresentationName(),
          'Failed to parse duration [{}]: {}.'.format(
              string, _SubException(e)))

  def Display(self, value):
    """Returns the display string for a duration value, leading PT dropped."""
    d = times.ParseDuration('{}s'.format(value))
    s = times.FormatDuration(d)
    if s.startswith('PT'):
      s = s[2:].lower()
    return s


class TimeStamp(TypeWithIntervalConstraint):
  """TimeStamp concept.

  Attributes:
    _fmt: string, a times.FormatDateTime() format specification.
    _string: bool, normalize value to a string if True
    _tzinfo: datetime.tzinfo, a time zone object, typically times.UTC or
      times.LOCAL.
  """

  def __init__(self, name, fmt=None, tz=None, string=False, **kwargs):
    self._fmt = fmt
    self._tzinfo = tz or times.UTC
    self._string = string
    super(TimeStamp, self).__init__(name, **kwargs)

  def BuildHelpText(self):
    """Appends timestamp syntax to the original help text."""
    return (
        '{}Must be a string representing an ISO 8601 date/time. Relative '
        'durations (prefixed by - or +) may be used to specify offsets from '
        'the current time. {}See $ gcloud topic datetimes for more '
        'information.'.format(
            _Insert(super(TimeStamp, self).BuildHelpText()),
            _Insert(self.Constraints())))

  def Convert(self, string):
    """Converts a datetime value from string returns it."""
    if not string:
      return None
    try:
      return times.ParseDateTime(string, tzinfo=self._tzinfo)
    except times.Error as e:
      raise exceptions.ParseError(
          self.GetPresentationName(),
          'Failed to parse duration [{}]: {}.'.format(
              string, _SubException(e)))

  def Display(self, value):
    """Returns the display string for a datetime value."""
    return times.FormatDateTime(value, fmt=self._fmt, tzinfo=self._tzinfo)

  def Normalize(self, value):
    return self.Display(value) if self._string else value


class SemVer(TypeWithIntervalConstraint):
  """SemVer concept."""

  def BuildHelpText(self):
    """Appends SemVer syntax to the original help text."""
    return (
        '{}Must be a string representing a SemVer number of the form '
        '_MAJOR_._MINOR_._PATCH_, where omitted trailing parts default to 0. '
        '{}See https://semver.org/ for more information.'.format(
            _Insert(super(SemVer, self).BuildHelpText()),
            _Insert(self.Constraints())))

  def Convert(self, string):
    """Converts a SemVer object from string returns it."""
    if not string:
      return None
    try:
      parts = string.split('.')
      while len(parts) < 3:
        parts.append('0')
      string = '.'.join(parts)
      return semver.SemVer(string)
    except semver.ParseError as e:
      raise exceptions.ParseError(self.GetPresentationName(), _SubException(e))

  def Display(self, value):
    """Returns the display string for a SemVer object value."""
    return '{}.{}.{}'.format(value.major, value.minor, value.patch)


class List(TypeWithSizeConstraint):
  """A list attribute concept.

  Attributes:
    _delim: string, the user specified lit item delimiter character.
    _element: string, the list element type object.
  """

  def __init__(self, name, element=None, **kwargs):
    super(List, self).__init__(name, constraint_kind='list length', **kwargs)
    self._delim = self._DEFAULT_DELIM
    self._element = element or String(name)

  def BuildHelpText(self):
    """Appends list syntax to the original help text."""
    item_help_text = self._element.BuildHelpText()
    if item_help_text.startswith('Must '):
      item_help_text = 'Each item m{}'.format(item_help_text[1:])
    elif item_help_text.startswith('A '):
      item_help_text = 'Each item is a{}'.format(item_help_text[1:])
    return (
        '{}Must be a string representing a list of `,` separated {} values. '
        '{}{}See $ gcloud topic escaping for details on using alternate '
        'delimiters.'.format(
            _Insert(super(List, self).BuildHelpText()),
            self._element.name,
            _Insert(self.Constraints()),
            _Insert(item_help_text)))

  def Display(self, value):
    """Returns the display string for the list value."""
    return self._delim.join([self._element.Display(item) for item in value])

  def Marshal(self):
    """Returns the list of concepts that this concept marshals."""
    return [self._element]

  def Parse(self, dependencies):
    """Parses the list from dependencies and returns it."""
    list_value = []
    try:
      value = dependencies.value
    except deps_lib.AttributeNotFoundError as e:
      if not self.required:
        return None
      raise exceptions.MissingRequiredArgumentError(
          self.GetPresentationName(), e)
    items, self._delim = self._Split(value)
    for item in items:
      # TODO(b/117144623): eventually use dependencies.marshalled_dependencies.
      item_dependencies = dependency_managers.DependencyViewFromValue(item)
      list_value.append(self._element.Parse(item_dependencies))
    list_value = self.Convert(list_value)
    self.Validate(list_value)
    return self.Normalize(list_value)


class Dict(TypeWithSizeConstraint):
  """A dict attribute concept.

  Attributes:
    _delim: string, the user specified lit item delimiter character.
    _entries: {string: ConceptType}, the map of concept type names to
      concept type objects.
    _additional: ConceptType, uninstantiated type used for unknown keys.
      If not specified then unknown keys are an error.
  """

  def __init__(self, name, entries=None, additional=None, **kwargs):
    super(Dict, self).__init__(
        name, constraint_kind='number of entries', **kwargs)
    self._delim = self._DEFAULT_DELIM
    self._entries = {}
    if entries:
      for entry in entries:
        self._entries[entry.name] = entry
    if not self._entries and not additional:
      additional = String
    self._additional = additional

  def BuildHelpText(self):
    """Appends dict syntax to the original help text."""
    buf = io.StringIO()
    buf.write('\n+\n')
    for key, entry in sorted(six.iteritems(self._entries)):
      buf.write('*{}*::: {}\n'.format(key, entry.BuildHelpText()))
    if self._additional:
      may_must = 'may'
      if self._entries:
        text = 'Additional _key_ names are allowed.'
      else:
        text = 'Any _key_ name is accepted.'
      value_help_text = self._additional('*').BuildHelpText()
      if value_help_text.startswith('Must '):
        value_help_text = 'Each _value_ m{}'.format(value_help_text[1:])
      elif value_help_text.startswith('A '):
        value_help_text = 'Each _value_ is a{}'.format(value_help_text[1:])
      buf.write('```*```::: {} {}\n'.format(text, value_help_text))
    else:
      may_must = 'must'
    return (
        '{}Must be a string representing a list of `,` separated '
        '_key_=_value_ pairs. {}_key_ {} be one of:{}'.format(
            _Insert(super(Dict, self).BuildHelpText()),
            _Insert(self.Constraints()),
            may_must,
            buf.getvalue()))

  def Display(self, value):
    """Returns the display string for the dict value."""
    return self._delim.join(sorted(['{}={}'.format(k, self._element.Display(v))
                                    for k, v in six.iteritems(value)]))

  def Marshal(self):
    """Returns the list of concepts that this concept marshals."""
    return self._entries.values()

  def Parse(self, dependencies):
    """Parses the dict from dependencies and returns it."""
    dict_value = {}
    try:
      value = dependencies.value
    except deps_lib.AttributeNotFoundError as e:
      if not self.required:
        return None
      raise exceptions.MissingRequiredArgumentError(
          self.GetPresentationName(), e)
    items, self._delim = self._Split(value)
    for item in items:
      key, val = item.split('=', 1)
      entry = self._entries.get(key)
      if not entry:
        if not self._additional:
          raise exceptions.ParseError(
              self.GetPresentationName(),
              'Unknown dictionary key [{}].'.format(key))
        entry = self._additional(key)
      item_dependencies = copy.copy(dependencies.marshalled_dependencies.get(
          key))
      try:
        item_dependencies.value = val
      except AttributeError:
        item_dependencies = dependency_managers.DependencyViewFromValue(val)
      dict_value[key] = entry.Parse(item_dependencies)
    dict_value = self.Convert(dict_value)
    self.Validate(dict_value)
    return self.Normalize(dict_value)



# -*- coding: utf-8 -*- #
# Copyright 2016 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flags and helpers for the compute firewall-rules commands."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.command_lib.compute import completers as compute_completers
from googlecloudsdk.command_lib.compute import flags as compute_flags
from googlecloudsdk.command_lib.util.apis import arg_utils

# Needs to be indented to show up correctly in help text
LIST_WITH_ALL_FIELDS_FORMAT = """\
table(
                    name,
                    network,
                    direction,
                    priority,
                    sourceRanges.list():label=SRC_RANGES,
                    destinationRanges.list():label=DEST_RANGES,
                    allowed[].map().firewall_rule().list():label=ALLOW,
                    denied[].map().firewall_rule().list():label=DENY,
                    sourceTags.list():label=SRC_TAGS,
                    sourceServiceAccounts.list():label=SRC_SVC_ACCT,
                    targetTags.list():label=TARGET_TAGS,
                    targetServiceAccounts.list():label=TARGET_SVC_ACCT,
                    disabled
                )"""

DEFAULT_LIST_FORMAT = """\
    table(
      name,
      network.basename(),
      direction,
      priority,
      allowed[].map().firewall_rule().list():label=ALLOW,
      denied[].map().firewall_rule().list():label=DENY,
      disabled
    )"""

LIST_NOTICE = """\
To show all fields of the firewall, please show in JSON format: --format=json
To show all fields in table format, please see the examples in --help.
"""


class FirewallsCompleter(compute_completers.ListCommandCompleter):

  def __init__(self, **kwargs):
    super(FirewallsCompleter, self).__init__(
        collection='compute.firewalls',
        list_command='compute firewall-rules list --uri',
        **kwargs)


def FirewallRuleArgument(required=True, plural=False):
  return compute_flags.ResourceArgument(
      resource_name='firewall rule',
      completer=FirewallsCompleter,
      plural=plural,
      required=required,
      global_collection='compute.firewalls')


def AddEnableLogging(parser):
  parser.add_argument(
      '--enable-logging',
      action=arg_parsers.StoreTrueFalseAction,
      help="""\
      Enable logging for the firewall rule. Logs will be exported to
      StackDriver. Firewall logging is disabled by default. To enable logging
      for an existing rule, run:

        $ {command} MY-RULE --enable-logging

      To disable logging on an existing rule, run:

        $ {command} MY-RULE --no-enable-logging
      """)


def GetLoggingMetadataArg(messages):
  return arg_utils.ChoiceEnumMapper(
      '--logging-metadata',
      messages.FirewallLogConfig.MetadataValueValuesEnum,
      custom_mappings={
          'INCLUDE_ALL_METADATA': 'include-all',
          'EXCLUDE_ALL_METADATA': 'exclude-all'
      },
      help_str=('Adds or removes metadata fields to or from the reported '
                'firewall logs. Can only be specified if --enable-logging is '
                'true.'))


def AddLoggingMetadata(parser, messages):
  GetLoggingMetadataArg(messages).choice_arg.AddToParser(parser)



# -*- coding: utf-8 -*- #
# Copyright 2016 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Resource display info for the Calliope display module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.core.cache import cache_update_ops


class DisplayInfo(object):
  """Display info accumulator for priming Displayer.

  "legacy" logic will be dropped when the incremental Command class refactor
  is complete.

  NOTICE: If you add an attribute:
    (1) document it
    (2) handle it in AddLowerDisplayInfo()

  Attributes:
    cache_updater: A resource_cache.Updater class that will be instantiated
      and called to update the cache to reflect the resources returned by the
      calling command.
    filter: The default filter string. args.filter takes precedence.
    flatten: The default flatten string. args.flatten takes precedence.
    format: The default format string. args.format takes precedence.
    transforms: The filter/format transforms symbol dict.
    aliases: The resource name alias dict.
  """

  def __init__(self):
    self._cache_updater = None
    self._filter = None
    self._format = None
    self._flatten = None
    self._transforms = {}
    self._aliases = {}

  # pylint: disable=redefined-builtin, name matches args.format and --format
  def AddLowerDisplayInfo(self, display_info):
    """Add lower precedence display_info to the object.

    This method is called by calliope to propagate CLI low precedence parent
    info to its high precedence children.

    Args:
      display_info: The low precedence DisplayInfo object to add.
    """
    if not self._cache_updater:
      self._cache_updater = display_info.cache_updater
    if not self._filter:
      self._filter = display_info.filter
    if not self._format:
      self._format = display_info.format
    if not self._flatten:
      self._flatten = display_info.flatten
    if display_info.transforms:
      transforms = dict(display_info.transforms)
      transforms.update(self.transforms)
      self._transforms = transforms
    if display_info.aliases:
      aliases = dict(display_info.aliases)
      aliases.update(self._aliases)
      self._aliases = aliases

  def AddFormat(self, format):
    """Adds a format to the display info, newer info takes precedence.

    Args:
      format: The default format string. args.format takes precedence.
    """
    if format:
      self._format = format

  def AddFlatten(self, flatten):
    """Adds a flatten to the display info, newer info takes precedence.

    Args:
      flatten: [str], The list of flatten strings to use by default; e.g.
        ['fieldA.fieldB', 'fieldC.fieldD']. Flatten strings provided by
        args.flatten take precedence if the --flatten flag is specified.
    """
    if flatten:
      self._flatten = flatten

  def AddFilter(self, filter):
    """Adds a filter to the display info, newer info takes precedence.

    Args:
      filter: The default filter string. args.filter takes precedence.
    """
    if filter:
      self._filter = filter

  def AddTransforms(self, transforms):
    """Adds transforms to the display info, newer values takes precedence.

    Args:
      transforms: A filter/format transforms symbol dict.
    """
    if transforms:
      self._transforms.update(transforms)

  def AddUriFunc(self, uri_func):
    """Adds a uri transform to the display info using uri_func.

    Args:
      uri_func: func(resource), A function that returns the uri for a
        resource object.
    """
    def _TransformUri(resource, undefined=None):
      try:
        return uri_func(resource) or undefined
      except (AttributeError, TypeError):
        return undefined

    self.AddTransforms({'uri': _TransformUri})

  def AddAliases(self, aliases):
    """Adds aliases to the display info, newer values takes precedence.

    Args:
      aliases: {str: parsed_key} The resource name alias dict that maps short
        names to parsed keys. The parsed key for 'abc.xyz' is ['abc', 'xyz'].
        See the resource_lex.Lexer.Key() docstring for parsed key details.
    """
    if aliases:
      self._aliases.update(aliases)

  def AddCacheUpdater(self, cache_updater):
    """Adds a cache_updater to the display info, newer values takes precedence.

    The cache updater is called to update the resource cache for CreateCommand,
    DeleteCommand and ListCommand commands.

    Args:
      cache_updater: A resource_cache.Updater class that will be instantiated
        and called to update the cache to reflect the resources returned by the
        calling command. None disables cache update.
    """
    self._cache_updater = cache_updater or cache_update_ops.NoCacheUpdater

  @property
  def cache_updater(self):
    return self._cache_updater

  @property
  def format(self):
    return self._format

  @property
  def flatten(self):
    return self._flatten

  @property
  def filter(self):
    return self._filter

  @property
  def aliases(self):
    return self._aliases

  @property
  def transforms(self):
    return self._transforms



"""Generated client library for composer version v1alpha2."""
# NOTE: This file is autogenerated and should not be edited by hand.

from __future__ import absolute_import

from apitools.base.py import base_api
from googlecloudsdk.third_party.apis.composer.v1alpha2 import composer_v1alpha2_messages as messages


class ComposerV1alpha2(base_api.BaseApiClient):
  """Generated client library for service composer version v1alpha2."""

  MESSAGES_MODULE = messages
  BASE_URL = 'https://composer.googleapis.com/'
  MTLS_BASE_URL = 'https://composer.mtls.googleapis.com/'

  _PACKAGE = 'composer'
  _SCOPES = ['https://www.googleapis.com/auth/cloud-platform']
  _VERSION = 'v1alpha2'
  _CLIENT_ID = 'CLIENT_ID'
  _CLIENT_SECRET = 'CLIENT_SECRET'
  _USER_AGENT = 'google-cloud-sdk'
  _CLIENT_CLASS_NAME = 'ComposerV1alpha2'
  _URL_VERSION = 'v1alpha2'
  _API_KEY = None

  def __init__(self, url='', credentials=None,
               get_credentials=True, http=None, model=None,
               log_request=False, log_response=False,
               credentials_args=None, default_global_params=None,
               additional_http_headers=None, response_encoding=None):
    """Create a new composer handle."""
    url = url or self.BASE_URL
    super(ComposerV1alpha2, self).__init__(
        url, credentials=credentials,
        get_credentials=get_credentials, http=http, model=model,
        log_request=log_request, log_response=log_response,
        credentials_args=credentials_args,
        default_global_params=default_global_params,
        additional_http_headers=additional_http_headers,
        response_encoding=response_encoding)
    self.projects_locations_environments_dags_dagRuns_taskInstances = self.ProjectsLocationsEnvironmentsDagsDagRunsTaskInstancesService(self)
    self.projects_locations_environments_dags_dagRuns = self.ProjectsLocationsEnvironmentsDagsDagRunsService(self)
    self.projects_locations_environments_dags_tasks = self.ProjectsLocationsEnvironmentsDagsTasksService(self)
    self.projects_locations_environments_dags = self.ProjectsLocationsEnvironmentsDagsService(self)
    self.projects_locations_environments = self.ProjectsLocationsEnvironmentsService(self)
    self.projects_locations_imageVersions = self.ProjectsLocationsImageVersionsService(self)
    self.projects_locations_operations = self.ProjectsLocationsOperationsService(self)
    self.projects_locations = self.ProjectsLocationsService(self)
    self.projects = self.ProjectsService(self)

  class ProjectsLocationsEnvironmentsDagsDagRunsTaskInstancesService(base_api.BaseApiService):
    """Service class for the projects_locations_environments_dags_dagRuns_taskInstances resource."""

    _NAME = 'projects_locations_environments_dags_dagRuns_taskInstances'

    def __init__(self, client):
      super(ComposerV1alpha2.ProjectsLocationsEnvironmentsDagsDagRunsTaskInstancesService, self).__init__(client)
      self._upload_configs = {
          }

    def List(self, request, global_params=None):
      r"""Lists task instances for a specified DAG run.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsDagsDagRunsTaskInstancesListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListTaskInstancesResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}/dags/{dagsId}/dagRuns/{dagRunsId}/taskInstances',
        http_method='GET',
        method_id='composer.projects.locations.environments.dags.dagRuns.taskInstances.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['filter', 'pageSize', 'pageToken'],
        relative_path='v1alpha2/{+parent}/taskInstances',
        request_field='',
        request_type_name='ComposerProjectsLocationsEnvironmentsDagsDagRunsTaskInstancesListRequest',
        response_type_name='ListTaskInstancesResponse',
        supports_download=False,
    )

  class ProjectsLocationsEnvironmentsDagsDagRunsService(base_api.BaseApiService):
    """Service class for the projects_locations_environments_dags_dagRuns resource."""

    _NAME = 'projects_locations_environments_dags_dagRuns'

    def __init__(self, client):
      super(ComposerV1alpha2.ProjectsLocationsEnvironmentsDagsDagRunsService, self).__init__(client)
      self._upload_configs = {
          }

    def Get(self, request, global_params=None):
      r"""Retrieves a DAG run.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsDagsDagRunsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (DagRun) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}/dags/{dagsId}/dagRuns/{dagRunsId}',
        http_method='GET',
        method_id='composer.projects.locations.environments.dags.dagRuns.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha2/{+name}',
        request_field='',
        request_type_name='ComposerProjectsLocationsEnvironmentsDagsDagRunsGetRequest',
        response_type_name='DagRun',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists DAG runs of a DAG.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsDagsDagRunsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListDagRunsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}/dags/{dagsId}/dagRuns',
        http_method='GET',
        method_id='composer.projects.locations.environments.dags.dagRuns.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['filter', 'pageSize', 'pageToken'],
        relative_path='v1alpha2/{+parent}/dagRuns',
        request_field='',
        request_type_name='ComposerProjectsLocationsEnvironmentsDagsDagRunsListRequest',
        response_type_name='ListDagRunsResponse',
        supports_download=False,
    )

  class ProjectsLocationsEnvironmentsDagsTasksService(base_api.BaseApiService):
    """Service class for the projects_locations_environments_dags_tasks resource."""

    _NAME = 'projects_locations_environments_dags_tasks'

    def __init__(self, client):
      super(ComposerV1alpha2.ProjectsLocationsEnvironmentsDagsTasksService, self).__init__(client)
      self._upload_configs = {
          }

    def List(self, request, global_params=None):
      r"""Lists tasks of a DAG.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsDagsTasksListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListTasksResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}/dags/{dagsId}/tasks',
        http_method='GET',
        method_id='composer.projects.locations.environments.dags.tasks.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['pageSize', 'pageToken'],
        relative_path='v1alpha2/{+parent}/tasks',
        request_field='',
        request_type_name='ComposerProjectsLocationsEnvironmentsDagsTasksListRequest',
        response_type_name='ListTasksResponse',
        supports_download=False,
    )

  class ProjectsLocationsEnvironmentsDagsService(base_api.BaseApiService):
    """Service class for the projects_locations_environments_dags resource."""

    _NAME = 'projects_locations_environments_dags'

    def __init__(self, client):
      super(ComposerV1alpha2.ProjectsLocationsEnvironmentsDagsService, self).__init__(client)
      self._upload_configs = {
          }

    def Get(self, request, global_params=None):
      r"""Retrieves a DAG.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsDagsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Dag) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}/dags/{dagsId}',
        http_method='GET',
        method_id='composer.projects.locations.environments.dags.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha2/{+name}',
        request_field='',
        request_type_name='ComposerProjectsLocationsEnvironmentsDagsGetRequest',
        response_type_name='Dag',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists DAGs in an environment.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsDagsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListDagsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}/dags',
        http_method='GET',
        method_id='composer.projects.locations.environments.dags.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['pageSize', 'pageToken'],
        relative_path='v1alpha2/{+parent}/dags',
        request_field='',
        request_type_name='ComposerProjectsLocationsEnvironmentsDagsListRequest',
        response_type_name='ListDagsResponse',
        supports_download=False,
    )

    def ListStats(self, request, global_params=None):
      r"""List DAGs with statistics for a given time interval.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsDagsListStatsRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListDagStatsResponse) The response message.
      """
      config = self.GetMethodConfig('ListStats')
      return self._RunMethod(
          config, request, global_params=global_params)

    ListStats.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}/dags:listStats',
        http_method='GET',
        method_id='composer.projects.locations.environments.dags.listStats',
        ordered_params=['environment'],
        path_params=['environment'],
        query_params=['interval_endTime', 'interval_startTime', 'pageSize', 'pageToken'],
        relative_path='v1alpha2/{+environment}/dags:listStats',
        request_field='',
        request_type_name='ComposerProjectsLocationsEnvironmentsDagsListStatsRequest',
        response_type_name='ListDagStatsResponse',
        supports_download=False,
    )

    def Trigger(self, request, global_params=None):
      r"""Trigger a DAG run.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsDagsTriggerRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (DagRun) The response message.
      """
      config = self.GetMethodConfig('Trigger')
      return self._RunMethod(
          config, request, global_params=global_params)

    Trigger.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}/dags/{dagsId}:trigger',
        http_method='POST',
        method_id='composer.projects.locations.environments.dags.trigger',
        ordered_params=['dag'],
        path_params=['dag'],
        query_params=[],
        relative_path='v1alpha2/{+dag}:trigger',
        request_field='triggerDagRequest',
        request_type_name='ComposerProjectsLocationsEnvironmentsDagsTriggerRequest',
        response_type_name='DagRun',
        supports_download=False,
    )

  class ProjectsLocationsEnvironmentsService(base_api.BaseApiService):
    """Service class for the projects_locations_environments resource."""

    _NAME = 'projects_locations_environments'

    def __init__(self, client):
      super(ComposerV1alpha2.ProjectsLocationsEnvironmentsService, self).__init__(client)
      self._upload_configs = {
          }

    def CheckUpgrade(self, request, global_params=None):
      r"""Check if an upgrade operation on the environment will succeed. In case of problems detailed info can be found in the returned Operation.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsCheckUpgradeRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('CheckUpgrade')
      return self._RunMethod(
          config, request, global_params=global_params)

    CheckUpgrade.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}:checkUpgrade',
        http_method='POST',
        method_id='composer.projects.locations.environments.checkUpgrade',
        ordered_params=['environment'],
        path_params=['environment'],
        query_params=[],
        relative_path='v1alpha2/{+environment}:checkUpgrade',
        request_field='checkUpgradeRequest',
        request_type_name='ComposerProjectsLocationsEnvironmentsCheckUpgradeRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def Create(self, request, global_params=None):
      r"""Create a new environment.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsCreateRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Create')
      return self._RunMethod(
          config, request, global_params=global_params)

    Create.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments',
        http_method='POST',
        method_id='composer.projects.locations.environments.create',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=[],
        relative_path='v1alpha2/{+parent}/environments',
        request_field='environment',
        request_type_name='ComposerProjectsLocationsEnvironmentsCreateRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Delete an environment.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}',
        http_method='DELETE',
        method_id='composer.projects.locations.environments.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha2/{+name}',
        request_field='',
        request_type_name='ComposerProjectsLocationsEnvironmentsDeleteRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Get an existing environment.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Environment) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}',
        http_method='GET',
        method_id='composer.projects.locations.environments.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha2/{+name}',
        request_field='',
        request_type_name='ComposerProjectsLocationsEnvironmentsGetRequest',
        response_type_name='Environment',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""List environments.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListEnvironmentsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments',
        http_method='GET',
        method_id='composer.projects.locations.environments.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['pageSize', 'pageToken'],
        relative_path='v1alpha2/{+parent}/environments',
        request_field='',
        request_type_name='ComposerProjectsLocationsEnvironmentsListRequest',
        response_type_name='ListEnvironmentsResponse',
        supports_download=False,
    )

    def ListImportErrors(self, request, global_params=None):
      r"""Lists DAG import errors.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsListImportErrorsRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListImportErrorsResponse) The response message.
      """
      config = self.GetMethodConfig('ListImportErrors')
      return self._RunMethod(
          config, request, global_params=global_params)

    ListImportErrors.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}:listImportErrors',
        http_method='GET',
        method_id='composer.projects.locations.environments.listImportErrors',
        ordered_params=['environment'],
        path_params=['environment'],
        query_params=['pageSize', 'pageToken'],
        relative_path='v1alpha2/{+environment}:listImportErrors',
        request_field='',
        request_type_name='ComposerProjectsLocationsEnvironmentsListImportErrorsRequest',
        response_type_name='ListImportErrorsResponse',
        supports_download=False,
    )

    def LoadSnapshot(self, request, global_params=None):
      r"""Loads a snapshot of a Cloud Composer environment. As a result of this operation, a snapshot of environment's specified in LoadSnapshotRequest is loaded into the environment.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsLoadSnapshotRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('LoadSnapshot')
      return self._RunMethod(
          config, request, global_params=global_params)

    LoadSnapshot.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}:loadSnapshot',
        http_method='POST',
        method_id='composer.projects.locations.environments.loadSnapshot',
        ordered_params=['environment'],
        path_params=['environment'],
        query_params=[],
        relative_path='v1alpha2/{+environment}:loadSnapshot',
        request_field='loadSnapshotRequest',
        request_type_name='ComposerProjectsLocationsEnvironmentsLoadSnapshotRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def Patch(self, request, global_params=None):
      r"""Update an environment.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsPatchRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Patch')
      return self._RunMethod(
          config, request, global_params=global_params)

    Patch.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}',
        http_method='PATCH',
        method_id='composer.projects.locations.environments.patch',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['updateMask'],
        relative_path='v1alpha2/{+name}',
        request_field='environment',
        request_type_name='ComposerProjectsLocationsEnvironmentsPatchRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def RestartWebServer(self, request, global_params=None):
      r"""Restart Airflow web server.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsRestartWebServerRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('RestartWebServer')
      return self._RunMethod(
          config, request, global_params=global_params)

    RestartWebServer.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}:restartWebServer',
        http_method='POST',
        method_id='composer.projects.locations.environments.restartWebServer',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha2/{+name}:restartWebServer',
        request_field='restartWebServerRequest',
        request_type_name='ComposerProjectsLocationsEnvironmentsRestartWebServerRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def SaveSnapshot(self, request, global_params=None):
      r"""Creates a snapshots of a Cloud Composer environment. As a result of this operation, snapshot of environment's state is stored in a location specified in the SaveSnapshotRequest.

      Args:
        request: (ComposerProjectsLocationsEnvironmentsSaveSnapshotRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('SaveSnapshot')
      return self._RunMethod(
          config, request, global_params=global_params)

    SaveSnapshot.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/environments/{environmentsId}:saveSnapshot',
        http_method='POST',
        method_id='composer.projects.locations.environments.saveSnapshot',
        ordered_params=['environment'],
        path_params=['environment'],
        query_params=[],
        relative_path='v1alpha2/{+environment}:saveSnapshot',
        request_field='saveSnapshotRequest',
        request_type_name='ComposerProjectsLocationsEnvironmentsSaveSnapshotRequest',
        response_type_name='Operation',
        supports_download=False,
    )

  class ProjectsLocationsImageVersionsService(base_api.BaseApiService):
    """Service class for the projects_locations_imageVersions resource."""

    _NAME = 'projects_locations_imageVersions'

    def __init__(self, client):
      super(ComposerV1alpha2.ProjectsLocationsImageVersionsService, self).__init__(client)
      self._upload_configs = {
          }

    def List(self, request, global_params=None):
      r"""List ImageVersions for provided location.

      Args:
        request: (ComposerProjectsLocationsImageVersionsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListImageVersionsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/imageVersions',
        http_method='GET',
        method_id='composer.projects.locations.imageVersions.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['includePastReleases', 'pageSize', 'pageToken'],
        relative_path='v1alpha2/{+parent}/imageVersions',
        request_field='',
        request_type_name='ComposerProjectsLocationsImageVersionsListRequest',
        response_type_name='ListImageVersionsResponse',
        supports_download=False,
    )

  class ProjectsLocationsOperationsService(base_api.BaseApiService):
    """Service class for the projects_locations_operations resource."""

    _NAME = 'projects_locations_operations'

    def __init__(self, client):
      super(ComposerV1alpha2.ProjectsLocationsOperationsService, self).__init__(client)
      self._upload_configs = {
          }

    def Cancel(self, request, global_params=None):
      r"""Starts asynchronous cancellation on a long-running operation. The server makes a best effort to cancel the operation, but success is not guaranteed. If the server doesn't support this method, it returns `google.rpc.Code.UNIMPLEMENTED`. Clients can use Operations.GetOperation or other methods to check whether the cancellation succeeded or whether the operation completed despite cancellation. On successful cancellation, the operation is not deleted; instead, it becomes an operation with an Operation.error value with a google.rpc.Status.code of 1, corresponding to `Code.CANCELLED`.

      Args:
        request: (ComposerProjectsLocationsOperationsCancelRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Cancel')
      return self._RunMethod(
          config, request, global_params=global_params)

    Cancel.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/operations/{operationsId}:cancel',
        http_method='POST',
        method_id='composer.projects.locations.operations.cancel',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha2/{+name}:cancel',
        request_field='cancelOperationRequest',
        request_type_name='ComposerProjectsLocationsOperationsCancelRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes a long-running operation. This method indicates that the client is no longer interested in the operation result. It does not cancel the operation. If the server doesn't support this method, it returns `google.rpc.Code.UNIMPLEMENTED`.

      Args:
        request: (ComposerProjectsLocationsOperationsDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/operations/{operationsId}',
        http_method='DELETE',
        method_id='composer.projects.locations.operations.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha2/{+name}',
        request_field='',
        request_type_name='ComposerProjectsLocationsOperationsDeleteRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets the latest state of a long-running operation. Clients can use this method to poll the operation result at intervals as recommended by the API service.

      Args:
        request: (ComposerProjectsLocationsOperationsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Operation) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/operations/{operationsId}',
        http_method='GET',
        method_id='composer.projects.locations.operations.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1alpha2/{+name}',
        request_field='',
        request_type_name='ComposerProjectsLocationsOperationsGetRequest',
        response_type_name='Operation',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists operations that match the specified filter in the request. If the server doesn't support this method, it returns `UNIMPLEMENTED`. NOTE: the `name` binding allows API services to override the binding to use different resource name schemes, such as `users/*/operations`. To override the binding, API services can add a binding such as `"/v1/{name=users/*}/operations"` to their service configuration. For backwards compatibility, the default name includes the operations collection id, however overriding users must ensure the name binding is the parent resource, without the operations collection id.

      Args:
        request: (ComposerProjectsLocationsOperationsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListOperationsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1alpha2/projects/{projectsId}/locations/{locationsId}/operations',
        http_method='GET',
        method_id='composer.projects.locations.operations.list',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['filter', 'pageSize', 'pageToken'],
        relative_path='v1alpha2/{+name}/operations',
        request_field='',
        request_type_name='ComposerProjectsLocationsOperationsListRequest',
        response_type_name='ListOperationsResponse',
        supports_download=False,
    )

  class ProjectsLocationsService(base_api.BaseApiService):
    """Service class for the projects_locations resource."""

    _NAME = 'projects_locations'

    def __init__(self, client):
      super(ComposerV1alpha2.ProjectsLocationsService, self).__init__(client)
      self._upload_configs = {
          }

  class ProjectsService(base_api.BaseApiService):
    """Service class for the projects resource."""

    _NAME = 'projects'

    def __init__(self, client):
      super(ComposerV1alpha2.ProjectsService, self).__init__(client)
      self._upload_configs = {
          }



# -*- coding: utf-8 -*- #
# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Packet mirroring."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.core import exceptions
from googlecloudsdk.core import log


class PacketMirroring(object):
  """Abstracts PacketMirroring resource."""

  def __init__(self, ref, compute_client=None, registry=None):
    self.ref = ref
    self._compute_client = compute_client
    self._registry = registry

  @property
  def _client(self):
    return self._compute_client.apitools_client

  @property
  def _messages(self):
    return self._compute_client.messages

  def _MakeCreateRequestTuple(self, packet_mirroring):
    return (self._client.packetMirrorings, 'Insert',
            self._messages.ComputePacketMirroringsInsertRequest(
                project=self.ref.project,
                region=self.ref.region,
                packetMirroring=packet_mirroring))

  def Create(self,
             packet_mirroring=None,
             only_generate_request=False,
             is_async=False):
    """Sends requests to create the packet mirroring."""
    requests = [self._MakeCreateRequestTuple(packet_mirroring)]

    if only_generate_request:
      return requests

    if not is_async:
      return self._compute_client.MakeRequests(requests)

    errors_to_collect = []
    result = self._compute_client.BatchRequests(requests, errors_to_collect)[0]

    if errors_to_collect:
      raise exceptions.MultiError(errors_to_collect)

    operation_ref = self._registry.Parse(
        result.name,
        collection='compute.regionOperations',
        params={
            'project': self.ref.project,
            'region': self.ref.region,
        },
        api_version='v1')
    log.CreatedResource(
        operation_ref,
        kind='packet mirroring [{0}]'.format(self.ref.Name()),
        is_async=True,
        details='Run the [gcloud compute operations describe] command '
        'to check the status of this operation.')
    return result

  def _MakeUpdateRequestTuple(self, packet_mirroring):
    return (self._client.packetMirrorings, 'Patch',
            self._messages.ComputePacketMirroringsPatchRequest(
                project=self.ref.project,
                region=self.ref.region,
                packetMirroring=self.ref.Name(),
                packetMirroringResource=packet_mirroring))

  def Update(self,
             packet_mirroring=None,
             only_generate_request=False,
             is_async=False,
             cleared_fields=None):
    """Sends requests to update the packet mirroring."""
    requests = [self._MakeUpdateRequestTuple(packet_mirroring)]

    if only_generate_request:
      return requests

    errors_to_collect = []

    with self._client.IncludeFields(cleared_fields or []):
      if not is_async:
        return self._compute_client.MakeRequests(requests)
      result = self._compute_client.BatchRequests(requests,
                                                  errors_to_collect)[0]

    if errors_to_collect:
      raise exceptions.MultiError(errors_to_collect)

    operation_ref = self._registry.Parse(
        result.name,
        collection='compute.regionOperations',
        params={
            'project': self.ref.project,
            'region': self.ref.region,
        },
        api_version='v1')
    log.UpdatedResource(
        operation_ref,
        kind='packet mirroring [{0}]'.format(self.ref.Name()),
        is_async=True,
        details='Run the [gcloud compute operations describe] command '
        'to check the status of this operation.')
    return result

  def MakeDeleteRequestTuple(self):
    return (self._client.packetMirrorings, 'Delete',
            self._messages.ComputePacketMirroringsDeleteRequest(
                region=self.ref.region,
                project=self.ref.project,
                packetMirroring=self.ref.Name()))

  def Delete(self, only_generate_request=False):
    requests = [self.MakeDeleteRequestTuple()]
    if not only_generate_request:
      return self._compute_client.MakeRequests(requests)
    return requests

  def _MakeDescribeRequestTuple(self):
    return (self._client.packetMirrorings, 'Get',
            self._messages.ComputePacketMirroringsGetRequest(
                region=self.ref.region,
                project=self.ref.project,
                packetMirroring=self.ref.Name()))

  def Describe(self, only_generate_request=False):
    requests = [self._MakeDescribeRequestTuple()]
    if not only_generate_request:
      return self._compute_client.MakeRequests(requests)
    return requests
